# GloFlow application and media management/publishing platform
# Copyright (C) 2021 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import argparse
import json

#--------------------------------------------------
def main():
	

	args_map = parse_args()



	contract_addr_str = args_map["contract_addr_str"]
	assert not contract_addr_str == None


	
	geth__port_int = 8545
	geth__host_str = "127.0.0.1"


	print(f"contract_addr - {contract_addr_str}")



	out_map = {}
	print(f"GF_OUT:{json.dumps(out_map)}")


#--------------------------------------------------
def parse_args():
	arg_parser = argparse.ArgumentParser(formatter_class = argparse.RawTextHelpFormatter)
	#----------------------------
	# CONTRACT_ADDRESS
	arg_parser.add_argument("-contract_addr", action = "store", default=None,
		help = "address of the target contract")

	#----------------------------
	cli_args_lst   = sys.argv[1:]
	args_namespace = arg_parser.parse_args(cli_args_lst)

	return {
		"contract_addr_str": args_namespace.contract_addr
	}

#--------------------------------------------------
main()


# GloFlow application and media management/publishing platform
# Copyright (C) 2023 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from collections import Counter

from PIL import Image
import numpy as np
from scipy.cluster.vq import kmeans, vq
import webcolors

#----------------------------------------------
# RUN
# gets the color palette for an image by using external libs

def run(p_image_path_str,
    p_num_clusters=3):
    
    image = Image.open(p_image_path_str)
    image = image.resize((150, 150)) # resize the image for faster processing

    # force convert image to RGB image always, in case its a RGBA
    image = image.convert('RGB')

    # convert image to RGB array
    image_array = np.array(image)
    image_array = image_array.reshape((image_array.shape[0] * image_array.shape[1], 3))

    # FIND_CLUSTERS
    centroids, _ = kmeans(image_array.astype(float), p_num_clusters)

    # VECTOR_QUANTIZATION
    # vq() - assigns each observation in a given dataset to the nearest centroid of the clusters
    #   identified by k-means; it maps each data point to the nearest cluster center.
    # cluster - array of indices indicating the closest centroid for each observation in the dataset.
    cluster, _ = vq(image_array, centroids)

    # count frequencies of clusters, to determine which cluster is the most dominant,
    # as a value of the RGB of each pixel.
    cluster_count = Counter(cluster)

    # sort clusters according to their frequency
    dominant_clusters = cluster_count.most_common(p_num_clusters)

    # get the colors for the top clusters
    top_colors = [tuple(centroids[cluster_idx].astype(int)) for cluster_idx, _ in dominant_clusters]



    


    return top_colors

#----------------------------------------------
def get_color_name(p_rgb_color):
    try:
        return webcolors.rgb_to_name(p_rgb_color)
    except ValueError:
        return closest_color(p_rgb_color)

#----------------------------------------------
def closest_color(p_color):

    delta_to_css3_colors_map = {}
    for key, name_str in webcolors.CSS3_HEX_TO_NAMES.items():
        
        r_c, g_c, b_c = webcolors.hex_to_rgb(key)
        
        rd = (r_c - p_color[0]) ** 2
        gd = (g_c - p_color[1]) ** 2
        bd = (b_c - p_color[2]) ** 2

        delta_to_css3_colors_map[(rd + gd + bd)] = name_str

    color_name_str = delta_to_css3_colors_map[min(delta_to_css3_colors_map.keys())]

    return color_name_str




# GloFlow application and media management/publishing platform
# Copyright (C) 2022 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import urllib.parse
import requests

#--------------------------------------------------
# RUN
def run(p_notify_completion_url_str,
	p_git_commit_hash_str = None,
	p_app_name_str        = None):
	
	url_str = None

	# add git_commit_hash as a querystring argument to the notify_completion URL.
	# the entity thats receiving the completion notification needs to know what the tag
	# is of the newly created container.
	if not p_git_commit_hash_str == None:
		
		url = urllib.parse.urlparse(p_notify_completion_url_str)
		
		# QUERY_STRING
		qs_lst = urllib.parse.parse_qsl(url.query)
		qs_lst.append(("base_img_tag", p_git_commit_hash_str)) # .parse_qs() places all values in lists

		qs_str = "&".join(["%s=%s"%(k, v) for k, v in qs_lst])

		# _replace() - "url" is of type ParseResult which is a subclass of namedtuple;
		#              _replace is a namedtuple method that:
		#              "returns a new instance of the named tuple replacing
		#              specified fields with new values".
		url_new = url._replace(query=qs_str)
		url_str = url_new.geturl()
	else:
		url_str = p_notify_completion_url_str

	print("NOTIFY_COMPLETION - HTTP REQUEST - %s"%(url_str))
	print(f"GIT commit_hash - {p_git_commit_hash_str}")

	#--------------------------
	# HTTP_POST

	data_map = {}
	if not p_app_name_str == None:
		data_map["app_name"] = p_app_name_str

	r = requests.post(url_str, json=data_map)
	print(r.text)

	if not r.status_code == 200:
		print("notify_completion http request failed")
		exit(1)
	
	#--------------------------


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

from colored import fg, bg, attr
import delegator

sys.path.append("%s/../../gf_core"%(modd_str))
import gf_core_cli

#--------------------------------------------------
# RUN_IN_CONTAINER
def run_in_cont(p_name_str,
    p_go_dir_path_str,
    p_go_output_path_str,
    p_static_bool    = False,
    p_local_dev_bool = False,
    p_repo_local_path_str = os.path.abspath(f'{modd_str}/../../../../../gloflow').strip()):
    assert isinstance(p_local_dev_bool, bool)

    cmd_lst = [
        "sudo", "docker", "run",
        "--rm", # remove after exit
    ]

    # FIX!! - run "go env GOMODCACHE" in build container as well to dynamically determine
    #         where its module cache dir is, to avoid this hardcoded value (in case it changes)
    #         and this code gets broken then.
    host_go_module_cache_dir_str            = delegator.run("go env GOMODCACHE").out.strip()
    build_container_go_module_cache_dir_str = "/root/go/pkg/mod"

    # VOLUMES
    volumes_lst = [
        # mount GF repo into the container
        "-v", f"{p_repo_local_path_str}:/home/gf",

        #-----------------------------
        # IMPORTANT!!
        # Go module cache - mount the host users go module cache dir into the
        #                   build containers go module cache dir. this will drastically speed up builds
        #                   on dev local machines since it will not download modules each time.
        "-v", f"{host_go_module_cache_dir_str}:{build_container_go_module_cache_dir_str}"

        #-----------------------------
    ]

    cmd_lst.extend(volumes_lst)

    cmd_lst.extend([

        "glofloworg/gf_builder_go_ubuntu:latest",

        # "python3 -u", "/home/gf/build/gf_builder/gf_builder.py", "-run=build_go"
        "python3", "-u", "/home/gf/ops/cli__build.py", "-run=build_go", "-build_outof_cont", f"-app={p_name_str}"
    ])

    p = gf_core_cli.run__view_realtime(cmd_lst, {},
        "gf_build_go", "green")

    p.wait()

#--------------------------------------------------
# RUN
def run(p_name_str,
    p_go_dir_path_str,
    p_go_output_path_str,
    p_static_bool       = False,
    p_exit_on_fail_bool = True,
    p_dynamic_libs_dir_path_str = os.path.abspath("%s/../../../../rust/build"%(modd_str)),
    p_go_get_bool = False):
    assert isinstance(p_static_bool, bool)
    
    print("")
    if p_static_bool:
        print(" -- %sSTATIC BINARY BUILD%s"%(fg("yellow"), attr(0)))
        
    print(" -- build %s%s%s service"%(fg("green"), p_name_str, attr(0)))
    print(" -- go_dir_path    - %s%s%s"%(fg("green"), p_go_dir_path_str, attr(0)))  
    print(" -- go_output_path - %s%s%s"%(fg("green"), p_go_output_path_str, attr(0)))  

    assert os.path.isdir(p_go_dir_path_str)
    assert os.path.isdir(os.path.dirname(p_go_output_path_str))

    print("--------------------")
    print(f"GO ENV\n")
    gf_core_cli.run(f"go env")
    print("--------------------")

    cwd_str = os.getcwd()
    os.chdir(p_go_dir_path_str) # change into the target main package dir

    #--------------------------------------------------
    def get_libs_for_linking():

        print("--------------------")
        print("LIBS_LINKING\n")

        # RUST_DYNAMIC_LIBS
        dynamic_libs_dir_path_str    = os.path.abspath(f"{modd_str}/../../../../rust/build")
        tf_dynamic_libs_dir_path_str = os.path.abspath(f"{modd_str}/../../../../rust/build/tf_lib/lib")

        print(f"dynamic libs dir - {fg('green')}{dynamic_libs_dir_path_str}{attr(0)}")
        gf_core_cli.run(f"ls -al {dynamic_libs_dir_path_str}")


        LD_paths_lst = [
            dynamic_libs_dir_path_str,
            tf_dynamic_libs_dir_path_str
        ]
        LD_paths_str = f"LD_LIBRARY_PATH={':'.join(LD_paths_lst)}"

        print("--------------------")

        return LD_paths_str

    #--------------------------------------------------
    LD_paths_str = get_libs_for_linking()

    # GO_GET
    # DEPRECATED!! - on modern Golang version "go build" will download modules as well,
    #                so there is no need for a "go get" to be run.
    if p_go_get_bool:
        _, _, exit_code_int = gf_core_cli.run(f"{LD_paths_str} go get -u") 
        print("")
        print("")
    
    #-----------------------------
    # STATIC_LINKING - when deploying to containers it is not always guaranteed that all
    #                  required libraries are present. so its safest to compile to a statically
    #                  linked lib.
    #                  build time a few times larger then regular, so slow for dev.
    # "-ldflags '-s'" - omit the symbol table and debug information.

    c_str = None
    if p_static_bool:
        
        print("--------------------")
        print(f"{fg('yellow')}STATIC LINKING{attr(0)}\n")
        
        # https://golang.org/cmd/link/
        # IMPORTANT!! - "CGO_ENABLED=0" and "-installsuffix cgo" no longer necessary since golang 1.10.
        #               "CGO_ENABLED=0" we also dont want to disable since Rust libs are used in Go via CGO.
        
        # IMPORTANT!! - debug .a files:
        #   "ar -t libgf_images_jobs.a" - get a list of Archived object files in static .a libs.
        #                                 static library is an archive (ar) of object files.
        #                                 The object files are usually in the ELF format

        gf_core_cli.run(f"ldconfig -v")
        # gf_core_cli.run(f"cp {dynamic_libs_dir_path_str}/libgf_images_jobs.a /usr/lib")

        args_lst = [
            
            LD_paths_str,
            # f"LD_LIBRARY_PATH={dynamic_libs_dir_path_str}",
            # f"LD_LIBRARY_PATH=/usr/lib",

            # "CGO_ENABLED=0",
            "GOOS=linux",
            "go build",

            # force rebuilding of packages that are already up-to-date.
            "-a",
            
            #-----------------------------
            # IMPORTANT!! - important to build apps with this buildtag, which enables files
            #   that import C libs that are compiled from the GF Rust codebase into a binary dynamic lib.
            #   without this tag that file is excluded, and golang tests can execute without being able
            #   to dynamically link with Rust libs.
            "-tags withrust",
            
            #-----------------------------
            
            # "-installsuffix cgo",

            # LINKER_FLAGS
            # "-ldflags"    - arguments to pass on each go tool link invocation
            # "-s"          - Omit the symbol table and debug information
            # "-extldflags" - Set space-separated flags to pass to the external linker.
            #                 on Alpine builds the GCC toolchain linker "ld" is used.
            # "-static"     - On systems that support dynamic linking, this 
            #                 overrides -pie and prevents linking with the shared libraries.
            # "-ldl"        - "-l" provides lib path. links in  /usr/lib/libdl.so/.a
            #                 this is needed to prevent Rust .a lib errors relating
            #                 to undefined references to "dlsym","dladdr"
            #
            # (f'''-ldflags '-s -extldflags "-t -static -lgf_images_jobs -ldl -lglib"' ''').strip(),
            # (f'''-ldflags '-s -extldflags "-lm"' ''').strip(),
            ('''-ldflags '-s -extldflags "-static -ldl"' ''').strip(),
            
            "-o %s"%(p_go_output_path_str),
        ]
        c_str = " ".join(args_lst)
    
    #-----------------------------
    # DYNAMIC_LINKING - fast build for dev.
    else:
        print("--------------------")
        print(f"{fg('yellow')}DYNAMIC LINKING{attr(0)}\n")

        #-----------------------------
        # IMPORTANT!! - important to build apps with this buildtag, which enables files
        #   that import C libs that are compiled from the GF Rust codebase into a binary dynamic lib.
        #   without this tag that file is excluded, and golang tests can execute without being able
        #   to dynamically link with Rust libs.
        "-tags withrust",
        
        #-----------------------------
        c_str = f"{LD_paths_str} go build -tags withrust -buildvcs=false -o {p_go_output_path_str}"

    #-----------------------------
    
    _, _, exit_code_int = gf_core_cli.run(c_str)

    # IMPORTANT!! - if "go build" returns a non-zero exit code in some environments (CI) we
    #               want to fail with a non-zero exit code as well - this way other CI 
    #               programs will flag builds as failed.
    if not exit_code_int == 0:
        if p_exit_on_fail_bool:
            exit(exit_code_int)

    os.chdir(cwd_str) # return to initial dir

    print("--------------------")
    print("build done...")


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import pymongo

#----------------------------------------------
# ADD!! - figure out a smarter way to pick the right hostport from p_host_port_lst,
#         instead of just picking the first element

def get_client(p_log_fun, p_host_port_lst = ['127.0.0.1:27017']):

	host_str,port_str = p_host_port_lst[0].split(':')

	mongo_client = pymongo.MongoClient(host_str, int(port_str))
	return mongo_client


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

sys.path.append("%s/../gf_core"%(modd_str))
import gf_core_cli


#----------------------------------------------
# JOBS
#----------------------------------------------
def job_run(p_config__file_path_str,
	p_acl_token_secret_id_str,
	p_host_url_str                   = "127.0.0.1:4646",
	p_ca_intermediate__file_path_str = None,
	p_cert_combined__file_path_str   = None,
	p_cert_key__file_path_str        = None,
	p_sudo_bool                      = False):

	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"NOMAD_TOKEN='%s'"%(p_acl_token_secret_id_str),
		"nomad job run",
		"-address=https://%s"%(p_host_url_str)
	])

	if not p_ca_intermediate__file_path_str == None:
		cmd_lst.extend([
			# path to a PEM encoded CA cert file to use to verify the Nomad server SSL certificate.
			"-ca-cert=%s"%(p_ca_intermediate__file_path_str),
			
			# ath to a PEM encoded client certificate for TLS authentication to the Nomad server.
			"-client-cert=%s"%(p_cert_combined__file_path_str),
			"-client-key=%s"%(p_cert_key__file_path_str),
		])

	cmd_lst.extend([
		p_config__file_path_str
	])

	stdout_str, _, return_code = gf_core_cli.run_cmd(" ".join(cmd_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

#----------------------------------------------
def job_status(p_name_str,
	p_acl_token_secret_id_str,
	p_host_url_str                   = "127.0.0.1:4646",
	p_ca_intermediate__file_path_str = None,
	p_cert_combined__file_path_str   = None,
	p_cert_key__file_path_str        = None,
	p_sudo_bool                      = False):

	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"NOMAD_TOKEN='%s'"%(p_acl_token_secret_id_str),
		"nomad job status",
		"-address=https://%s"%(p_host_url_str)
	])
	
	if not p_ca_intermediate__file_path_str == None:
		cmd_lst.extend([
			# path to a PEM encoded CA cert file to use to verify the Nomad server SSL certificate.
			"-ca-cert=%s"%(p_ca_intermediate__file_path_str),
			
			# ath to a PEM encoded client certificate for TLS authentication to the Nomad server.
			"-client-cert=%s"%(p_cert_combined__file_path_str),
			"-client-key=%s"%(p_cert_key__file_path_str),
		])

	stdout_str, _, return_code = gf_core_cli.run_cmd(" ".join(cmd_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

#----------------------------------------------
# ACL
#----------------------------------------------
# ACL_BOOTSTRAP
def acl_bootstrap(p_host_url_str = "127.0.0.1:4646",
	p_ca_intermediate__file_path_str = None,
	p_cert_combined__file_path_str   = None,
	p_cert_key__file_path_str        = None,
	p_sudo_bool                      = False):

	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"nomad acl bootstrap"
		"-address=https://%s"%(p_host_url_str)
	])

	if not p_ca_intermediate__file_path_str == None:
		cmd_lst.extend([
			# path to a PEM encoded CA cert file to use to verify the Nomad server SSL certificate.
			"-ca-cert=%s"%(p_ca_intermediate__file_path_str),
			
			# ath to a PEM encoded client certificate for TLS authentication to the Nomad server.
			"-client-cert=%s"%(p_cert_combined__file_path_str),
			"-client-key=%s"%(p_cert_key__file_path_str),
		])

	stdout_str, _, return_code = gf_core_cli.run_cmd(" ".join(cmd_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

#----------------------------------------------
# ACL_TOKEN_CREATE
def acl_token_create(p_name_str,
	p_output_file_path_str,
	p_acl_token_secret_id_str,
	p_policies_lst                   = [],
	p_type_str                       = "client",
	p_host_url_str                   = "127.0.0.1:4646",
	p_ca_intermediate__file_path_str = None,
	p_cert_combined__file_path_str   = None,
	p_cert_key__file_path_str        = None,
	p_sudo_bool = False):
	assert p_type_str == "management" or \
		p_type_str == "client"

	# "nomad acl token self" - get information about the current token

	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"NOMAD_TOKEN='%s'"%(p_acl_token_secret_id_str),
		"nomad acl token create",

		"-name='%s'"%(p_name_str),
		"-type='%s'"%(p_type_str),

		# GLOBAL_TOKEN - are created in the authoritative region and then replicate to all other regions
		"-global",

		"-address=https://%s"%(p_host_url_str)
	])

	if not p_ca_intermediate__file_path_str == None:
		cmd_lst.extend([
			# path to a PEM encoded CA cert file to use to verify the Nomad server SSL certificate.
			"-ca-cert=%s"%(p_ca_intermediate__file_path_str),
			
			# ath to a PEM encoded client certificate for TLS authentication to the Nomad server.
			"-client-cert=%s"%(p_cert_combined__file_path_str),
			"-client-key=%s"%(p_cert_key__file_path_str),
		])

	for p in p_policies_lst:
		cmd_lst.append("-policy='%s'"%(p))

	stdout_str, _, return_code = gf_core_cli.run_cmd(" ".join(cmd_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

	#-------------
	# WRITE_TO_FILE

	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")
	fs_write_cmd_str = '''bash -c "echo '%s' > %s"'''%(stdout_str, p_output_file_path_str)
	cmd_lst.append(fs_write_cmd_str)

	_, _, return_code = gf_core_cli.run_cmd(" ".join(cmd_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

	#-------------

#----------------------------------------------
# ACL_POLICY_APPLY
def acl_policy_apply(p_name_str,
	p_description_str,
	p_config__file_path_str,
	p_acl_token_secret_id_str,
	p_host_url_str                   = "127.0.0.1:4646",
	p_ca_intermediate__file_path_str = None,
	p_cert_combined__file_path_str   = None,
	p_cert_key__file_path_str        = None,
	p_sudo_bool = False):
	
	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"NOMAD_TOKEN='%s'"%(p_acl_token_secret_id_str),
		"nomad acl policy apply",
		"-description '%s'"%(p_description_str),
		"-address=https://%s"%(p_host_url_str),
	])

	if not p_ca_intermediate__file_path_str == None:
		cmd_lst.extend([
			# path to a PEM encoded CA cert file to use to verify the Nomad server SSL certificate.
			"-ca-cert=%s"%(p_ca_intermediate__file_path_str),
			
			# path to a PEM encoded client certificate for TLS authentication to the Nomad server.
			"-client-cert=%s"%(p_cert_combined__file_path_str),
			"-client-key=%s"%(p_cert_key__file_path_str),
		])

	cmd_lst.extend([
		p_name_str,
		p_config__file_path_str
	])

	stdout_str, _, return_code = gf_core_cli.run_cmd(" ".join(cmd_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

#----------------------------------------------
# ACL_POLICY_LIST
def acl_policy_list(p_acl_token_secret_id_str,
	p_host_url_str = "127.0.0.1:4646",
	p_ca_intermediate__file_path_str = None,
	p_cert_combined__file_path_str   = None,
	p_cert_key__file_path_str        = None,
	p_sudo_bool = False):

	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"NOMAD_TOKEN='%s'"%(p_acl_token_secret_id_str),
		"nomad acl policy list",
		"-address=https://%s"%(p_host_url_str),
	])

	if not p_ca_intermediate__file_path_str == None:
		cmd_lst.extend([
			# path to a PEM encoded CA cert file to use to verify the Nomad server SSL certificate.
			"-ca-cert=%s"%(p_ca_intermediate__file_path_str),
			
			# ath to a PEM encoded client certificate for TLS authentication to the Nomad server.
			"-client-cert=%s"%(p_cert_combined__file_path_str),
			"-client-key=%s"%(p_cert_key__file_path_str),
		])

	stdout_str, _, return_code = gf_core_cli.run_cmd(" ".join(cmd_lst), p_env_map = None)
	if not return_code == 0:
		print("CLI failed...")
		exit()

	policies_lst = []
	for l in stdout_str.strip().split("\n")[1:]:

		policy_name_str = l.split()[0].strip()
		policies_lst.append({
			"name_str": policy_name_str
		})
	return policies_lst


from gf_core import gf_core_db_sql, gf_core_id, gf_core_error, gf_core_logger
from gf_apps.gf_images.gf_images_client import gf_images_client
from gf_apps.gf_images.gf_images_core   import gf_image_db_sql, gf_image

from gf_observe import gf_extern_load
from gf_ml import gf_llm_core
from gf_extern_services.gf_aws import gf_aws_ec2, gf_aws_s3, gf_aws_secrets, gf_aws_route53
from gf_extern_services.gf_google import gf_google_sheets

print("gloflow...")
version = "0.1.18"

#-------------------------
# EXPORT_API
# simplified unified GF Py API with flattened namespace

# DB
class db():
    init         = gf_core_db_sql.init_db_client
    table_exists = gf_core_db_sql.table_exists

# CORE
class core():
    create_id    = gf_core_id.create
    create_error = gf_core_error.create
    get_log_fun  = gf_core_logger.get_log_fun

# IMAGES
class images():
    load_adt         = gf_image.load_adt
    add              = gf_images_client.add_image
    put_images_in_db = gf_image_db_sql.put_images

# OBSERVE
class observe():
    init           = gf_extern_load.init
    ext_load       = gf_extern_load.observe
    get_cached     = gf_extern_load.get_cached
    get_cached_group = gf_extern_load.get_cached_group
    relate           = gf_extern_load.relate

# ML
class ml():
    run_llm = gf_llm_core.run_llm

# EXTERN SERVICES
class extern():
    aws_ec2 = gf_aws_ec2
    aws_s3  = gf_aws_s3
    aws_secrets = gf_aws_secrets
    aws_route53 = gf_aws_route53
    gcp_sheets  = gf_google_sheets


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from colored import fg, attr
import time

#----------------------------------------------
def get_log_fun():

    def log_color_display_fun(p_group, p_msg):
        t = str(time.time())
        
        if p_group == 'FUN_ENTER':
            print(f"{t}:{fg('yellow')}FUN_ENTER{attr('reset')}:{p_msg}")
        
        elif p_group == 'ERROR':
            print(f"{t}:{fg('red')}{p_group}{attr('reset')}:{p_msg}")
        
        elif p_group == 'INFO':
            print(f"{t}:{fg('green')}{p_group}{attr('reset')}:{fg('green')}{p_msg}{attr('reset')}")
        
        elif p_group == 'INFO_USR':
            print(f"{t}:{fg('magenta')}{p_group}{attr('reset')}:{fg('magenta')}{p_msg}{attr('reset')}")

		# log message is in some way related to external-systems data
        elif p_group == 'EXTERN':
            print(f"{t}:{fg('blue')}{p_group}{attr('reset')}:{p_msg}")

		# if 'TEST' is anywhere in the group string
		# elif p_group.find('TEST'):
		#	clint.textui.puts(clint.textui.colored.cyan(p_group)+':'+p_msg)
        
		#--------------
		# JAVASCRIPT FORMATING
        elif p_group == 'JS:FUN_ENTER':
            print(f"{t}:       {fg('yellow')}FUN_ENTER{attr('reset')}:{p_msg}")
        
        elif p_group == 'JS:INFO':
            print(f"{t}:       {fg('green')}{p_group}{attr('reset')}:{fg('green')}{p_msg}{attr('reset')}")

		#--------------
            
    return log_color_display_fun


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from colored import fg, bg, attr

#---------------------------------------------------
def log_fun(g, m):
    if g == "ERROR":
        print("%s%s%s:%s%s%s"%(bg("red"), g, attr(0), fg("red"), m, attr(0)))
    else:
        print("%s%s%s:%s%s%s"%(fg("yellow"), g, attr(0), fg("green"), m, attr(0)))


# GloFlow application and media management/publishing platform
# Copyright (C) 2025 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
from gf_core import gf_core_utils
from . import gf_image

#---------------------------------------------------
def eliminate_duplicate_tags(p_image_id_str,
	p_db_client):
	
	table_name_str = "gf_images"
	cur = p_db_client.cursor()

	query_str = f'''
		UPDATE {table_name_str}
		SET tags_lst = ARRAY(
			SELECT DISTINCT UNNEST(tags_lst)
		)
		WHERE id = %s;
		'''
	cur.execute(query_str, (p_image_id_str,))

#---------------------------------------------------
def has_duplicate_tags(p_image_id_str,
	p_db_client):

	table_name_str = "gf_images"
	cur = p_db_client.cursor()

	query_str = f'''
		SELECT EXISTS (
			SELECT 1 FROM {table_name_str}
			WHERE id = %s
			AND array_length(tags_lst, 1) > 
				array_length(ARRAY(SELECT DISTINCT UNNEST(tags_lst)), 1)
		);
	'''
	
	try:
		cur.execute(query_str, (p_image_id_str,))
		exists_bool = cur.fetchone()[0]
		cur.close()
		
		return exists_bool
	except Exception as e:
		print("Error:", e)
		return False  # Default to False if an error occurs

#---------------------------------------------------
def get_image(p_image_id_str,
	p_db_client):

	table_name_str = "gf_images"
	cur = p_db_client.cursor()

	query_str = f'''
		SELECT *
		FROM {table_name_str}
		WHERE id = %s
	'''

	cur.execute(query_str, (p_image_id_str,))

	row = cur.fetchone()
	if row is None:
		return None

	columns = [desc[0] for desc in cur.description]
	image_map = dict(zip(columns, row))

	cur.close()
	return image_map

#---------------------------------------------------
# p_mongo_source_bool - temporary, used during mongo->sql migration

def put_images(p_image_adt_lst,
	p_db_client,
	p_mongo_source_bool: bool = False):
	assert all(isinstance(img, gf_image.GFimage) for img in p_image_adt_lst), \
		"All items must be instances of gf_image.GFimage"

	table_name_str = "gf_images"
	cur = p_db_client.cursor()

	query_str = f'''INSERT INTO {table_name_str} (
			v,
			id,
			creation_time,
			user_id,

			client_type,
			title,
			flows_names,

			origin_url,
			origin_page_url,

			original_file_int_url,

			thumb_small_url,
			thumb_medium_url,
			thumb_large_url,

			format,
			width,
			height,

			dominant_color_hex,
			palette_colors_hex,

			meta_map,
			tags_lst,

			mongo_source
		)

		VALUES (
			%s, %s, %s, %s,
			%s, %s, %s,
			%s, %s,
			%s,
			%s, %s, %s,
			%s, %s, %s,
			%s, %s,
			%s, %s,
			%s
		)
	'''

	# prepare data for batch insertion
	insertion_data_lst = []
	for p_image_adt in p_image_adt_lst:
		
		sql_timestamp_str = gf_core_utils.unix_to_sql_timestamp(p_image_adt.creation_unix_time_f)
		insertion_data_lst.append((
			"0.1",  # v
			p_image_adt.id_str,
			sql_timestamp_str,
			p_image_adt.user_id_str,

			p_image_adt.client_type_str,
			p_image_adt.title_str,
			p_image_adt.flows_names_lst,

			p_image_adt.origin_url_str,
			p_image_adt.origin_page_url_str,

			p_image_adt.original_file_int_url_str,

			p_image_adt.thumb_small_url_str,
			p_image_adt.thumb_medium_url_str,
			p_image_adt.thumb_large_url_str,

			p_image_adt.format_str,
			p_image_adt.width_int,
			p_image_adt.height_int,

			p_image_adt.dominant_color_hex_str,
			p_image_adt.palette_colors_hex_lst,

			json.dumps(p_image_adt.meta_map),
			p_image_adt.tags_lst,

			p_mongo_source_bool
		))

	# batch insert
	cur.executemany(query_str, insertion_data_lst)
	
	p_db_client.commit()
	cur.close()

#---------------------------------------------------------------------------------
	
def check_images_exist(p_ids_lst, p_db_client):
	assert isinstance(p_ids_lst, list) and \
		all(isinstance(p, str) for p in p_ids_lst), "Input must be a list of strings"

	table_name_str = "gf_images"
	cur = p_db_client.cursor()


	# createa list with as many "%s" (sql var placeholder) as there are IDs in p_ids_lst
	# and then join it into a single string with ", " as separator
	ids_placeholder_str = ', '.join(['%s'] * len(p_ids_lst))

	query_str = f'''
		SELECT id
		FROM {table_name_str}
		WHERE id IN ({ids_placeholder_str})
	'''

	cur.execute(query_str, tuple(p_ids_lst))

	# get a list of ID's that already exist in the database	
	existing_ids_lst = cur.fetchall()
	existing_ids_set = set([row[0] for row in existing_ids_lst])

	# compose a list of bools, same length as p_ids_lst, that for each
	# ID indicates if its already present or not.
	exists_bool_lst = [id_str in existing_ids_set for id_str in p_ids_lst]

	cur.close()

	return exists_bool_lst


# GloFlow application and media management/publishing platform
# Copyright (C) 2021 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

from colored import fg, bg, attr

sys.path.append("%s/../utils"%(modd_str))
import gf_core_cli

#--------------------------------------------------
def run(p_test_ci_bool, p_aws_region_str):


	print(f"    {fg('green')}TEST SQS_PROCESS{attr(0)} >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
	
	
	# FIX!! - specify which SQS queue to consume a message from, this will use the default
	#         queue name, not the test queue.
	bin_str = f"{modd_str}/../../build/gf_web3_monitor"
	cmd_lst = [
		bin_str,
		"test", "worker_event_process",
		f"--config={modd_str}/../../config/gf_web3_monitor.yaml"
	]

	p = gf_core_cli.run__view_realtime(cmd_lst, {
			"AWS_REGION":            p_aws_region_str,
			"AWS_ACCESS_KEY_ID":     os.environ["AWS_ACCESS_KEY_ID"],
			"AWS_SECRET_ACCESS_KEY": os.environ["AWS_SECRET_ACCESS_KEY"],
			"GF_AWS_SQS_QUEUE":      os.environ["GF_AWS_SQS_QUEUE"]
		},
		"gf_web3_monitor", "green")


# GloFlow application and media management/publishing platform
# Copyright (C) 2024 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
extern_load 
	- helps with tracking and monitoring of external data loading.
	- tracks loads of external data as events in the DB.
	- allows for historic tracking of external data loading.
	- simple mechanism, py native.
	- results of loads are stored in the DB in string form
		- results can be html or json.
		- if json then stored in postgres jsonb type.
	- the partition key is stored for each load event.
		- parititon keys can be multidimensional
		- each dim separated by "__".
"""

import os, json
import boto3
from icecream import ic
import gloflow as gf

from gf_observe import gf_extern_load_db as db

# stores data on loading/processing of external models info
# such as from automani or some other source.
default_s3_bucket_name_str  = "gf"

#---------------------------------------------------------------------------------
# PUBLIC
#---------------------------------------------------------------------------------
# OBSERVE
# p_url_str - url from which the observation is coming, if relevant.
# p_resp_store_file_path_str - allow external users of this API to determine the file path
# 	                           to be used for caching the response data. user knows best what
#                              naming structure makes sense for their domain.
#                              files are stored in some block storage, local or remote (s3, etc.).
#                              filepath should be relative.
# p_related_observations_ids_lst - list of observation ids that this observation depends on.
#                                  that dependance is usually in the form of a data dependency.
# p_group_id_str - user can supply assign observations to a group, if needed.
#                  useful if multiple observations are related and need to be grouped.

def observe(p_load_type_str,
	p_part_key_str,
	p_source_domain_str,
	p_runtime_map,
	p_group_id_str        = None,
	p_meta_map            = {},
	p_url_str             = None,
	p_resp_data_map       = None,
	p_cache_file_path_str = None):
	
	assert(isinstance(p_load_type_str, str))

	if p_cache_file_path_str is None:
		store_result_bool = False
	else:
		store_result_bool = True

	#-----------------------
	# RESP_FILE_STORAGE
	cache_s3_key_str = None
	if store_result_bool:

		assert(isinstance(p_resp_data_map, dict))
		resp_data_str = json.dumps(p_resp_data_map)
			
		cache_s3_key_str = upload_cache(resp_data_str,
			p_load_type_str,
			p_cache_file_path_str,
			p_source_domain_str,
			p_runtime_map)

	#-----------------------
	# DB
	observation_id_str = db.insert(p_load_type_str,
		p_part_key_str,
		p_source_domain_str,
		cache_s3_key_str,
		p_runtime_map["db_client"],
		p_meta_map     = p_meta_map,
		p_url_str      = p_url_str,
		p_group_id_str = p_group_id_str)

	ic(observation_id_str)

	#-----------------------
	# MONITORING
	if "monitoring_fun" in p_runtime_map:
		assert(callable(p_runtime_map["monitoring_fun"]))
		p_runtime_map["monitoring_fun"](p_event_map={
				"type":  "gf.observe.extern_load",
				"level": "info",

				# APP_LEVEL
				"load_type":     p_load_type_str,
				"source_domain": p_source_domain_str,
			})

	#-----------------------
		
	return observation_id_str

#---------------------------------------------------------------------------------
# RELATE
def relate(p_observation_id_int,
	p_related_observations_ids_lst,
	p_runtime_map):
	
	assert(isinstance(p_observation_id_int, int))
	assert(isinstance(p_related_observations_ids_lst, list))

	print(f"relate observation {p_observation_id_int} to observations {p_related_observations_ids_lst}")
	
	db.relate_observations(p_observation_id_int,
		p_related_observations_ids_lst,
		p_runtime_map["db_client"])

#---------------------------------------------------------------------------------
# GET_CACHED

def get_cached(p_load_type_str,
	p_part_key_str,
	p_source_domain_str,
	p_runtime_map):
	assert(isinstance(p_load_type_str, str))
	assert(isinstance(p_part_key_str, str))
	assert(isinstance(p_runtime_map, dict))
	assert(isinstance(p_source_domain_str, str) or p_source_domain_str is None)

	#-----------------------
	# DB
	latest_load_map = db.get_latest(p_load_type_str,
		p_part_key_str,
		p_source_domain_str,
		p_runtime_map["db_client"])

	#-----------------------
	# S3
	
	ic(latest_load_map)
	
	bucket_name_str = p_runtime_map.get("s3_data_sink_bucket_str", default_s3_bucket_name_str)
	ic(bucket_name_str)

	s3 = boto3.client('s3')
	resp = s3.get_object(Bucket=bucket_name_str,
		Key=latest_load_map["resp_cache_file_path"])

	r = resp['Body']
	
	resp_str = resp['Body'].read().decode('utf-8')
	info_map = json.loads(resp_str)
	assert(isinstance(info_map, dict))

	#-----------------------
	
	observation_id_int = latest_load_map["id"]
	assert(isinstance(observation_id_int, int))

	return info_map, observation_id_int

#---------------------------------------------------------------------------------
# GET_CACHED_GROUP
def get_cached_group(p_load_type_str,
	p_part_key_str,
	p_source_domain_str,
	p_runtime_map):

	#-----------------------
	# DB
	latest_group_id_str = db.get_latest_group_id(p_load_type_str,
		p_source_domain_str,
		p_runtime_map["db_client"])

	#-----------------------

	ic(latest_group_id_str)



	loads_lst = db.get_group(p_load_type_str,
		p_part_key_str,
		p_source_domain_str,
		latest_group_id_str,
		p_runtime_map["db_client"])
	assert(isinstance(loads_lst, list))

	ic(loads_lst)


	#-----------------------
	# S3

	bucket_name_str = p_runtime_map.get("s3_data_sink_bucket_str", default_s3_bucket_name_str)
	ic(bucket_name_str)

	s3 = boto3.client('s3')

	#---------------------------------------------------------------------------------
	def get_file(p_load_map):
		
		resp = s3.get_object(Bucket=bucket_name_str,
			Key=p_load_map["resp_cache_file_path"])

		r = resp['Body']
		
		resp_str = resp['Body'].read().decode('utf-8')
		info_map = json.loads(resp_str)
		assert(isinstance(info_map, dict))
		return info_map
	
	#---------------------------------------------------------------------------------
	
	
	infos_lst = []
	for l_map in loads_lst:
		info_map = get_file(l_map)
		infos_lst.append(info_map)

	#-----------------------
	
	return infos_lst

#---------------------------------------------------------------------------------
def init(p_db_client):

	db.init(p_db_client)

#---------------------------------------------------------------------------------
# CACHE
#---------------------------------------------------------------------------------
def upload_cache(p_data_str,
	p_load_type_str,
	p_cache_file_path_str,
	p_source_domain_str,
	p_runtime_map):
	assert(isinstance(p_data_str, str))
	assert(isinstance(p_load_type_str, str))
	bucket_name_str = p_runtime_map.get("s3_data_sink_bucket_str", "gf")
	ic(bucket_name_str)

	s3 = boto3.client('s3')

	#-----------------------
	file_path_norm_str = os.path.normpath(p_cache_file_path_str)
	dir_str            = f"gf/ext_load/{p_source_domain_str}/{p_load_type_str}"
	s3_key_str         = f'{dir_str}/{file_path_norm_str}'
	ic(s3_key_str)

	#-----------------------
	# S3
	s3.put_object(Bucket=bucket_name_str,
		Key=s3_key_str,
		Body=p_data_str)

	#-----------------------

	return s3_key_str


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os

#--------------------------------------------------
# FILE

def get_from_file(p_aws_creds_file_path_str):
    print(p_aws_creds_file_path_str)
    assert os.path.isfile(os.path.abspath(p_aws_creds_file_path_str))

    f             = open(p_aws_creds_file_path_str, 'r')
    aws_creds_map = {}
    for l in f.readlines():

        if l == '' or l == '\n': continue
        if l.startswith('#'):    continue #ignore comments
        
        k, v = l.strip().split("=")
        k    = k.strip()
        v    = v.strip()

        if k == "AWS_ACCESS_KEY_ID" or \
            k == "AWS_SECRET_ACCESS_KEY":
            #k == "GF_AWS_TOKEN":
            aws_creds_map[k]=v
    f.close()   
    return aws_creds_map

#--------------------------------------------------
# ENV_VARS

def get_from_env_vars():
    aws_creds_map = {
        "aws_access_key_id_str": os.environ["AWS_ACCESS_KEY_ID"],
        "aws_secret_access_key": os.environ["AWS_SECRET_ACCESS_KEY"],
        #"aws_token_str": os.environ["GF_AWS_TOKEN"]
    }
    return aws_creds_map
    


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

sys.path.append("%s/../containers"%(modd_str))
import gf_os_docker
import gf_containers

sys.path.append("%s/../../meta"%(modd_str))
import gf_meta

#---------------------------------------------------

# DEPRECATED!! - not being used anymore
'''
def cont__publish(p_app_name_str,
	p_app_build_meta_map,
	p_gf_dockerhub_user_str,
	p_gf_dockerhub_pass_str,
	p_log_fun,
	p_git_commit_hash_str = None,
	p_docker_sudo_bool    = False):
	assert isinstance(p_gf_dockerhub_user_str, str)

	# PUBLISH
	gf_containers.publish(p_app_name_str,
		p_app_build_meta_map,
		p_gf_dockerhub_user_str,
		p_gf_dockerhub_pass_str,
		p_log_fun,
		p_git_commit_hash_str = p_git_commit_hash_str, 
		p_exit_on_fail_bool   = True,
		p_docker_sudo_bool    = p_docker_sudo_bool)
'''

#---------------------------------------------------
# DEPRECATED?? - is this still being used?

def cont__build(p_app_name_str,
	p_dockerhub_user_name_str,
	p_log_fun,
	p_docker_sudo_bool = False):
	assert isinstance(p_dockerhub_user_name_str, str)

	build_meta_map = gf_meta.get()["build_info_map"]
	assert p_app_name_str in build_meta_map.keys()

	gf_builder_meta_map            = build_meta_map[p_app_name_str]
	cont_image_name_str            = gf_builder_meta_map["cont_image_name_str"]
	cont_image_version_str         = gf_builder_meta_map["version_str"]
	cont_image_dockerfile_path_str = os.path.abspath(gf_builder_meta_map["dockerfile_path_str"])
	assert os.path.isfile(cont_image_dockerfile_path_str)

	# DOCKER_BUILD
	image_full_name_str = "%s/%s:%s"%(p_dockerhub_user_name_str,
		cont_image_name_str,
		cont_image_version_str)

	gf_os_docker.build_image([image_full_name_str],
		cont_image_dockerfile_path_str,
		p_log_fun,
		p_exit_on_fail_bool = True,
		p_docker_sudo_bool  = p_docker_sudo_bool)


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_dir = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))

import pprint
from gf_core import gf_core_error, gf_core_utils
from . import gf_image_utils

#---------------------------------------------------
def on_map(p_image_info_map,
	p_log_fun):
	assert isinstance(p_image_info_map, dict)
	
	# pprint.pprint(p_image_info_map)

	max_title_characters = 500
	#-------------------
	# 'id_str' - is None if image_info_dict comes from the outside of the system
	#            and the ID has not yet been assigned. 
	#            ID is determined if image_info_dict comes from the DB or someplace
	#            else within the system
	
	id_str = p_image_info_map.get('id_str', None)
	if id_str is not None:
		assert isinstance(id_str, str)
		assert id_str != ""

	#-------------------
	# CREATION TIME
	try:
		if 'creation_unix_time_f' in p_image_info_map.keys():
			assert isinstance(p_image_info_map['creation_unix_time_f'], float)

	except Exception as e:
		msg_str = '''creation_unix_time_f is invalid (not a float)'''
		gf_core_error.create(msg_str,
			"image_verification",
			{"id_str": id_str}, e, "gf_images",
			p_log_fun)
		raise Exception(msg_str)
	#-------------------
	# TITLE
	
	title_str = None
	try:
		assert 'title_str' in p_image_info_map.keys()
		title_str = p_image_info_map['title_str']
		
		assert isinstance(p_image_info_map['title_str'], str)
		assert len(p_image_info_map['title_str']) < max_title_characters

	except Exception as e:
		msg_str = '''title_str [%s] is missing, invalid, or has more then max_length [%s] characters'''%(title_str, max_title_characters)
		gf_core_error.create(msg_str,
			"image_verification",
			{"id_str": id_str}, e, "gf_images",
			p_log_fun)
		raise Exception(msg_str)

	#-------------------
	# FLOWS_NAMES

	try:

		assert 'flows_names_lst' in p_image_info_map.keys()
		flows_names_lst = p_image_info_map['flows_names_lst']
		assert isinstance(flows_names_lst, list)
		assert len(flows_names_lst) > 0

		for flow_name_str in flows_names_lst:
			assert isinstance(flow_name_str, str)
			assert flow_name_str != ""

	except Exception as e:
		msg_str = '''flows_names_lst [%s] is missing or is invalid'''
		gf_core_error.create(msg_str,
			"image_verification",
			{"id_str": id_str}, e, "gf_images",
			p_log_fun)
		raise Exception(msg_str)
	
	#-------------------
	# FORMAT 
	
	format_str = None
	try:
		assert "format_str" in p_image_info_map.keys()
		assert isinstance(p_image_info_map['format_str'], str)

		normalized_format_str = check_image_format(p_image_info_map['format_str'].lower(), p_log_fun)
		format_str = normalized_format_str
			
	except Exception as e:
		msg_str = '''format_str [%s] is missing or is invalid'''%(format_str)
		gf_core_error.create(msg_str,
			"image_verification",
			{"id_str": id_str}, e, "gf_images",
			p_log_fun,
			p_reraise_bool=True)
	
	#-------------------
	# WIDTH/HEIGHT
	
	try:
		assert 'width_int' in p_image_info_map.keys()
		width_int = p_image_info_map['width_int']
		
		assert 'height_int' in p_image_info_map.keys()
		height_int = p_image_info_map['height_int']
		
		assert isinstance(width_int, int)
		assert isinstance(height_int, int)
		       
	except Exception as e:
		msg_str = '''image width/height is missing or is invalid'''
		gf_core_error.create(msg_str,
			"image_verification",
			{"id_str": id_str}, e, "gf_images",
			p_log_fun,
			p_reraise_bool=True)
	
	#-------------------
	# ORIGIN URL
	
	origin_url_str = None
	try:
		assert 'origin_url_str' in p_image_info_map.keys()
		origin_url_str = p_image_info_map['origin_url_str']

		gf_core_utils.is_valid_url(origin_url_str)

	except Exception as e:
		msg_str = '''origin_url_str [%s] is missing or is invalid'''%(origin_url_str)
		gf_core_error.create(msg_str,
			"image_verification",
			{"id_str": id_str}, e, "gf_images",
			p_log_fun,
			p_reraise_bool=True)
	
	#-------------------
	# THUMBS
	thumbnail_small_url_str  = p_image_info_map.get('thumb_small_url_str',  None)
	thumbnail_medium_url_str = p_image_info_map.get('thumb_medium_url_str', None)
	thumbnail_large_url_str  = p_image_info_map.get('thumb_large_url_str',  None)

	#---------------------------------------------------
	def check_thumb(p_url_str):
		if p_url_str is not None:
			assert isinstance(p_url_str, str)
			assert p_url_str != ""

	#---------------------------------------------------
	
	check_thumb(thumbnail_small_url_str)
	check_thumb(thumbnail_medium_url_str)
	check_thumb(thumbnail_large_url_str)

	#-------------------

#---------------------------------------------------
def check_image_format(p_format_str, p_log_fun):
	
	normalized_format_str = None
	
	assert isinstance(p_format_str, str)
	assert p_format_str in gf_image_utils.IMAGE_EXTENSIONS	
				 
	# normalize "jpg" (variation on jpeg) to "jpeg"
	if p_format_str == 'jpg':
		normalized_format_str = 'jpeg'
	else:
		normalized_format_str = p_format_str
		
	return normalized_format_str


# GloFlow application and media management/publishing platform
# Copyright (C) 2023 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU G11eneral Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
from google.oauth2 import service_account
from apiclient import discovery
import pandas as pd

#---------------------------------------------------------------------------------
# NAMED_RANGES
#---------------------------------------------------------------------------------
# SET

def set_named_range(p_value_str,
	p_range_name_str,
	p_spreadsheet_id_str,
	p_service_client):
	"""
	Set a named range in a Google Sheet with a string value.

	Parameters:
	- p_value_str: The string value to set in the range.
	- p_range_name_str: The named range to set.
	- p_spreadsheet_id_str: The spreadsheet ID.
	- p_service_client: The authorized Google Sheets API client.
	"""

	body = {
		"range": p_range_name_str,
		"majorDimension": "COLUMNS",
		"values": [[p_value_str]] # single cell with the provided string value
	}

	# UPDATE
	p_service_client.spreadsheets().values().update(
		spreadsheetId=p_spreadsheet_id_str,
		range=p_range_name_str,

		# "USED_ENTERED" - processes the input as if it were entered directly into the cell by a user.
		#                  alternative is "RAW" - processes the input as if it were a literal string.
		valueInputOption='USER_ENTERED',
		body=body
	).execute()

#---------------------------------------------------------------------------------
# GET

def get_named_range(p_range_name_str,
	p_spreadsheet_id_str,
	p_service_client):


	result = p_service_client.spreadsheets().values().get(spreadsheetId=p_spreadsheet_id_str,

		range=p_range_name_str,

		# format returned data to be in column-first format
		majorDimension='COLUMNS').execute()

	# iterate over each column, and get the one thats needed.
	columns_vals_lst = []
	for column_vals_lst in result['values']:

		if len(column_vals_lst) > 0:
			columns_vals_lst.append(column_vals_lst)

	return columns_vals_lst

#---------------------------------------------------------------------------------
# ADD_NOTE_TO_NAMED_RANGE

def add_note_to_named_range(p_note_str,
	p_range_name_str,
	p_spreadsheet_id_str,
	p_service_client):
	"""
	Add a note to the first cell of a named range in a Google Sheet.

	Parameters:
	- p_note_str: The note to add to the named range.
	- p_range_name_str: The named range to add a note to.
	- p_spreadsheet_id_str: The spreadsheet ID.
	- p_service_client: The authorized Google Sheets API client.
	"""

	sheet_metadata = p_service_client.spreadsheets().get(
		spreadsheetId=p_spreadsheet_id_str
	).execute()

	ranges = sheet_metadata.get('namedRanges', [])
	target_range = next((r for r in ranges if r['name'] == p_range_name_str), None)

	if target_range:
		sheet_id_str = target_range['range']['sheetId']
		start_row = target_range['range']['startRowIndex']
		start_col = target_range['range']['startColumnIndex']

		# Set the note only for the first cell in the range
		note_body = {
			"requests": [
				{
					"updateCells": {
						"range": {
							"sheetId": sheet_id_str,
							"startRowIndex": start_row,
							"endRowIndex": start_row + 1,
							"startColumnIndex": start_col,
							"endColumnIndex": start_col + 1
						},
						"rows": [
							{
								"values": [
									{
										"note": p_note_str
									}
								]
							}
						],
						"fields": "note"
					}
				}
			]
		}

		p_service_client.spreadsheets().batchUpdate(
			spreadsheetId=p_spreadsheet_id_str,
			body=note_body
		).execute()
	else:
		raise ValueError("Named range not found.")

#---------------------------------------------------------------------------------
def add_hyperlink_above_named_range(p_url_str,
	p_range_name_str,
	p_spreadsheet_id_str,
	p_service_client,
	p_label_str="link"):
	"""
	Add a clickable hyperlink in the cell one row above the first cell of a named range in a Google Sheet.

	Parameters:
	- p_url_str: The URL to link to.
	- p_range_name_str: The named range to add the hyperlink above.
	- p_spreadsheet_id_str: The spreadsheet ID.
	- p_service_client: The authorized Google Sheets API client.
	- p_label_str: The label for the hyperlink (default is 'Click Here').
	"""

	hyperlink_formula_str = f'=HYPERLINK("{p_url_str}", "{p_label_str}")'

	sheet_metadata = p_service_client.spreadsheets().get(
		spreadsheetId=p_spreadsheet_id_str
	).execute()

	ranges = sheet_metadata.get('namedRanges', [])
	target_range = next((r for r in ranges if r['name'] == p_range_name_str), None)

	if target_range:

		sheet_id_int = target_range['range']['sheetId']
		start_row = target_range['range']['startRowIndex']
		start_col = target_range['range']['startColumnIndex']

		sheet_name_str = get_sheet_name(p_spreadsheet_id_str, sheet_id_int, p_service_client)

		col_letter = chr(65 + start_col) # Convert column index to letter (A, B, C...)
		cell_address_str = f"{sheet_name_str}!{col_letter}{start_row}"
		print(f"Adding hyperlink to cell: {cell_address_str}")

		p_service_client.spreadsheets().values().update(
			spreadsheetId=p_spreadsheet_id_str,
			range=cell_address_str,
			valueInputOption='USER_ENTERED',
			body={
				"range": cell_address_str,
				"majorDimension": "COLUMNS",
				"values": [[hyperlink_formula_str]]
			}
		).execute()
	else:
		raise ValueError("Named range not found.")

#---------------------------------------------------------------------------------
# VAR
#---------------------------------------------------------------------------------
def get_client(p_google_service_key_jobs_json_str):

	SCOPES = ['https://www.googleapis.com/auth/spreadsheets']

	print("google-sheets connecting...")
	credentials = service_account.Credentials.from_service_account_info(json.loads(p_google_service_key_jobs_json_str),
		scopes=SCOPES)
	service = discovery.build("sheets", "v4", credentials=credentials)
	return service

#---------------------------------------------------------------------------------
# IMPORTANT!! - downloads the whole sheet, all the values of columns

def get_all_columns(p_spreadsheet_id_str,
	p_subsheet_name_str,
	p_service_client):

	range_name_str = f"{p_subsheet_name_str}"

	result = p_service_client.spreadsheets().values().get(spreadsheetId=p_spreadsheet_id_str,
		# body=data,
		range=range_name_str,

		# format returned data to be in column-first format
		majorDimension='COLUMNS').execute()

	# iterate over each column, and get the one thats needed
	columns_vals_lst = []
	for column_vals_lst in result['values']:

		columns_vals_lst.append(column_vals_lst)

	return columns_vals_lst

#---------------------------------------------------------------------------------
def get_column_by_name_from_list(p_column_name_str,
	p_columns_vals_lst,
	p_column_name__row_index_int=1):
	assert isinstance(p_columns_vals_lst, list)

	# iterate over each column, and get the one thats needed
	i=0
	for column_vals_lst in p_columns_vals_lst:

		# first test if the column has values,
		# and if it has more values than the row number where the column name is stored
		if len(column_vals_lst) > 0 and len(column_vals_lst) > (p_column_name__row_index_int+1):

			column_name_str = column_vals_lst[p_column_name__row_index_int]
			if column_name_str == p_column_name_str:
				return column_vals_lst, i

		i+=1

	return None, 0

#---------------------------------------------------------------------------------
# block is a user-created sub-section of a sheet.

def get_columns_by_name_from_named_range(p_columns_names_lst,
	p_range_name_str,
	p_spreadsheet_id_str,
	p_service_client,
	p_column_name__row_index_int=0):


	columns_lst = get_named_range(p_range_name_str,
		p_spreadsheet_id_str,
		p_service_client)


	columns_vals_lst = []
	for column_name_str in p_columns_names_lst:

		i=0
		for vals_lst in columns_lst:

			if len(vals_lst) > 0 and len(vals_lst) > (p_column_name__row_index_int):

				a_column_name_str = vals_lst[p_column_name__row_index_int]

				if column_name_str == a_column_name_str:
					column_vals_lst = vals_lst
					columns_vals_lst.append((column_vals_lst, i))

					# column discovered, go to the next one
					break

			i+=1

	return columns_vals_lst

#---------------------------------------------------------------------------------
def get_column_by_name(p_column_name_str,
	p_spreadsheet_id_str,
	p_subsheet_name_str,
	p_service_client,

	# row in which the user-titles of columns are expected
	p_column_name__row_index_int=1):


	# get all values of all the columns in a sheet
	columns_vals_lst = get_all_columns(p_spreadsheet_id_str,
		p_subsheet_name_str,
		p_service_client)

	i=0

	# iterate over each column, and get the one thats needed
	for column_vals_lst in columns_vals_lst:

		# first test if the column has values,
		# and if it has more values than the row number where the column name is stored
		if len(column_vals_lst) > 0 and len(column_vals_lst) > (p_column_name__row_index_int+1):

			column_name_str = column_vals_lst[p_column_name__row_index_int]

			if column_name_str == p_column_name_str:
				return column_vals_lst, i

		i+=1

	return None, 0

#---------------------------------------------------------------------------------
def get_sheet_id(p_sheet_name_str,
	p_spreadsheet_id_str,
	p_service_client):

	spreadsheet_metadata = p_service_client.spreadsheets().get(spreadsheetId=p_spreadsheet_id_str).execute()
	sheets_lst           = spreadsheet_metadata.get('sheets', '')

	for sheet in sheets_lst:
		properties = sheet.get('properties', {})
		if properties.get('title', '').lower() == p_sheet_name_str.lower():
			sheet_id_int = properties.get('sheetId', '')
			return sheet_id_int

	return None

#---------------------------------------------------------------------------------
def get_sheet_name(p_spreadsheet_id_str,
	p_sheet_id_int,
	p_service_client):
	sheets_metadata = p_service_client.spreadsheets().get(spreadsheetId=p_spreadsheet_id_str).execute()
	for sheet in sheets_metadata['sheets']:
		if sheet['properties']['sheetId'] == p_sheet_id_int:
			return sheet['properties']['title']
	return None

#---------------------------------------------------------------------------------
def get_subcolumns_by_names(p_supercolumn_name_str,
	p_subcolumn_names_lst,
	p_spreadsheet_id_str,
	p_subsheet_name_str,
	p_service_client,
	p_supercolumn_row_index_int=0,
	p_subcolumn_row_index_int=1):
	"""
	Get multiple subcolumns by name that are located within a supercolumn.

	Parameters:
	- p_supercolumn_name_str: The name of the supercolumn (merged column)
	- p_subcolumn_names_lst: List of names of the subcolumns within the supercolumn
	- p_spreadsheet_id_str: The spreadsheet ID
	- p_subsheet_name_str: The name of the sheet
	- p_service_client: The authorized Google Sheets API client
	- p_supercolumn_row_index_int: Row index where the supercolumn name is located (default: 0)
	- p_subcolumn_row_index_int: Row index where the subcolumn name is located (default: 1)

	Returns:
	- Dictionary where keys are subcolumn names and values are dictionaries with "values" (column values)
	  and "column_index" (column index)
	"""
	assert isinstance(p_supercolumn_name_str, str), "p_supercolumn_name_str must be a string"
	assert isinstance(p_subcolumn_names_lst, list), "p_subcolumn_names_lst must be a list"
	assert isinstance(p_spreadsheet_id_str, str), "p_spreadsheet_id_str must be a string"
	assert isinstance(p_subsheet_name_str, str), "p_subsheet_name_str must be a string"

	# Get all columns from the sheet
	columns_vals_lst = get_all_columns(p_spreadsheet_id_str,
		p_subsheet_name_str,
		p_service_client)

	# Get merged cells information to handle supercolumns properly
	spreadsheet_metadata = p_service_client.spreadsheets().get(
		spreadsheetId=p_spreadsheet_id_str,
		includeGridData=False
	).execute()

	# Find the sheet ID
	sheet_id_int = None
	for sheet in spreadsheet_metadata.get('sheets', []):
		if sheet.get('properties', {}).get('title') == p_subsheet_name_str:
			sheet_id_int = sheet.get('properties', {}).get('sheetId')
			break

	if sheet_id_int is None:
		print(f"Sheet '{p_subsheet_name_str}' not found in spreadsheet '{p_spreadsheet_id_str}'.")
		return {}

	# Get the merged ranges for the target sheet.
	# Explanation:
	# In Google Sheets, when cells are merged, the API stores information about these merged regions
	# in the sheet's metadata under the "merges" key. Each merged range is represented as a dictionary
	# with start/end row and column indices. This is crucial for identifying supercolumns, because only
	# the top-left cell of a merged region contains the displayed value, while the rest are empty.
	# By retrieving these ranges, we can determine which columns are part of a merged supercolumn,
	# and thus correctly locate subcolumns within that supercolumn.
	merged_ranges_lst = []
	for sheet in spreadsheet_metadata.get('sheets', []):
		if sheet.get('properties', {}).get('sheetId') == sheet_id_int:
			merged_ranges_lst = sheet.get('merges', [])
			break

	# First, find the column that contains the supercolumn name
	# In merged cells, only the leftmost cell contains the value
	supercolumn_start_col_int = None
	i = 0
	for column_vals_lst in columns_vals_lst:
		if (len(column_vals_lst) > p_supercolumn_row_index_int and
			column_vals_lst[p_supercolumn_row_index_int] == p_supercolumn_name_str):
			supercolumn_start_col_int = i
			break
		i += 1

	# If we didn't find the supercolumn name, return empty dict
	if supercolumn_start_col_int is None:
		print(f"Supercolumn '{p_supercolumn_name_str}' not found in the specified row index.")
		return {}

	#---------------------------------------------------------------------------------
	def get_supercolumn_columns():

		# Find the merged range that contains this column at the supercolumn row
		# This will help identify all columns that are part of the merged supercolumn
		supercolumn_cols_lst = [supercolumn_start_col_int]  # Start with the column that has the name

		for merged_range in merged_ranges_lst:
			start_row_index = merged_range.get('startRowIndex', 0)
			end_row_index = merged_range.get('endRowIndex', 0)
			start_col_index = merged_range.get('startColumnIndex', 0)
			end_col_index = merged_range.get('endColumnIndex', 0)

			# Check if this merged range is at the supercolumn row and includes our starting column
			if (start_row_index <= p_supercolumn_row_index_int < end_row_index and
				start_col_index <= supercolumn_start_col_int < end_col_index):

				# This merged range represents our supercolumn
				# Add all columns in this range
				for col_idx in range(start_col_index, end_col_index):
					if col_idx not in supercolumn_cols_lst:
						supercolumn_cols_lst.append(col_idx)

		return supercolumn_cols_lst

	#---------------------------------------------------------------------------------
	supercolumn_cols_lst = get_supercolumn_columns()

	# Initialize an empty DataFrame with columns in the order of supercolumn_cols_lst
	results_df = pd.DataFrame()

	# Keep track of which subcolumn names we've found
	found_subcolumns_set = set()

	# Now, within the identified supercolumn columns, find all specified subcolumns
	for col_index_int in supercolumn_cols_lst:

		if col_index_int >= len(columns_vals_lst):
			print(f"Column index {col_index_int} is out of range for the columns list.")
			continue

		column_vals_lst = columns_vals_lst[col_index_int]

		if len(column_vals_lst) <= p_subcolumn_row_index_int:
			print(f"Column {col_index_int} does not have enough rows to contain subcolumn names.")
			continue

		subcolumn_name_str = column_vals_lst[p_subcolumn_row_index_int]

		if subcolumn_name_str in p_subcolumn_names_lst:
			# Extract only the values after the header rows
			# Add 1 to p_subcolumn_row_index_int to start from the row after the subcolumn name
			data_start_index = p_subcolumn_row_index_int + 1
			data_values_lst = column_vals_lst[data_start_index:] if data_start_index < len(column_vals_lst) else []

			# Add the column to the DataFrame
			results_df[subcolumn_name_str] = pd.Series(data_values_lst)

			# Mark this subcolumn as found
			found_subcolumns_set.add(subcolumn_name_str)

			# If we've found all requested subcolumns, we can stop searching
			if len(found_subcolumns_set) == len(p_subcolumn_names_lst):
				break

	# Reorder the DataFrame columns to match the order in supercolumn_cols_lst
	results_df = results_df[[name for name in p_subcolumn_names_lst if name in results_df.columns]]

	return results_df



# GloFlow application and media management/publishing platform
# Copyright (C) 2023 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
from urllib.parse import urlparse
from pathlib import Path
from PIL import Image
import requests

IMAGE_EXTENSIONS = ['jpg', 'jpeg', 'png', 'gif', 'bmp', 'webp', 'tiff']

#--------------------------------------------
def get_image_metadata(p_image_path_str):
	assert(os.path.isfile(p_image_path_str))
	
	with Image.open(p_image_path_str) as img:
		width, height = img.size
		img_type_str = img.format

	meta_map = {
		"width_int":  width,
		"height_int": height,
		"format_str": img_type_str
	}

	return meta_map

#--------------------------------------------
def check_image_format(p_url_str):
	
	# do a simple file-path check if its an image
	is_image_bool, format_str = is_image_url(p_url_str)
	if is_image_bool:
		return True, format_str
	
	# if the extension doesn't indicate an image, check the Content-Type using a GET request
	try:
		# using HEAD request to minimize data transfer
		response = requests.head(p_url_str)
		if 'image' in response.headers.get('Content-Type', '').lower():
			
			# check the format based on the Content-Type
			content_type_str = response.headers['Content-Type'].split('/')[1]
			return True, content_type_str
		
	except requests.exceptions.RequestException:
		return False, None
	
	return False, None

#--------------------------------------------
def is_image_url(p_url_str):
	
	path = urlparse(p_url_str).path
	
	# check if the URL ends with a valid image file extension
	ext_str = Path(path).suffix.lower().strip(".")
	if ext_str in IMAGE_EXTENSIONS:
		
		format_str = ext_str
		return True, format_str
	else:
		return False, None


# GloFlow application and media management/publishing platform
# Copyright (C) 2021 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import argparse
import json
from colored import fg, bg, attr
import numpy as np
from PIL import Image, ImageDraw

#----------------------------------------------
def run_from_file(p_image_file_path_str):
	assert os.path.isfile(p_image_file_path_str)

	image = Image.open(p_image_file_path_str)
	width_int, height_int = image.size

	palette_pixels_only_arr, result = run(image, width_int, height_int)

	return palette_pixels_only_arr, result

#----------------------------------------------
def run(p_image,
	p_img_width_int,
	p_img_height_int):

	img_pixels_original_arr = np.array(p_image.getdata())

	# only select the first 3 channels of each pixel, since some images have
	# an alpha channel as well which we dont want.
	img_pixels_arr = img_pixels_original_arr[:, :3]

	#----------------------------------------------
	# levels_max gives a color palette of certain size:
	# 3 - 8  colors
	# 4 - 16 colors
	# 5 - 32 colors
	# 6 - 64 colors

	def process_img__with_px_coords(p_img_pixels_with_global_index_arr,
		p_level_int,
		p_levels_max_int=3):
		
		# TERMINATION
		# check if we reached the maximum level of subdivision/recursion 
		if p_level_int >= p_levels_max_int:
			

			img_pixels_global_indexes_arr = p_img_pixels_with_global_index_arr[:, 0]

  
			# for all rows, only include last 3 elements (rgb pixel values), since thats what
			# we want to calculate the average with.
			img_pixels_no_global_indexes_arr = p_img_pixels_with_global_index_arr[:, 1:4]

			# get average pixel, by taking averages of all pixels on each of the channels.
			# this is the last image subdivision (dividing by pixel channel with maximum range), 
			# and so calculate average pixel value for that subdivision and return.
			# 
			# img_pixels_no_global_indexes_arr.T - average is calculated on a transpose of the pixels matrix,
			#                                      so that each channel is an array, and average is
			#                                      calculated on each channel. 
			average_pixel_arr = np.average(img_pixels_no_global_indexes_arr.T, axis=1)

			# round and cast as integer
			a = np.round(average_pixel_arr).astype(int)

			return [(img_pixels_global_indexes_arr,
				a[0],  # R 
				a[1],  # G
				a[2])] # B

		# remove the pixels global index (index of the pixel relative to the whole image)
		# 
		# input:
		# p_img_pixels_with_global_index_arr
		# [[     0      1      2 ... 115997 115998 115999]
		# [   162    163    164 ...    142    140    140]
		# [   200    201    202 ...    193    191    191]
		# [   203    204    205 ...    194    195    195]]
		# 
		# [1:] - exclude the the first element/row of the transposed matrix,
		#        since its the pixel_global_index and we dont need ranges on that.
		# 
		# output:
		# [   162    163    164 ...    142    140    140]
		# [   200    201    202 ...    193    191    191]
		# [   203    204    205 ...    194    195    195]]
		img_by_channel_arr = p_img_pixels_with_global_index_arr.T[1:]

		#----------
		# PICK_CHANNEL_WITH_MAX_RANGE
		# get the range (max_val-min_val) of each of the r/g/b channels.
		# .ptp() - stands for "peak-to-peak" and it calculates the range of values in a NumPy array along a specified axis.
		#          returns the difference between the maximum and minimum values in the array along the specified axis.
		ranges_per_channel_lst = img_by_channel_arr.ptp(axis=1)

		# argmax() - returns the indices of the maximum values along a specified axis of a NumPy array.
		channel_with_max_range_index_int = np.argmax(ranges_per_channel_lst)

		# values of pixels in a particular channel which has the maximum range of values.
		#
		# img_by_channel_arr+1 - to account for the fact that the first element is global_index,
		#                        so the color index has to be offset by 1.
		img_max_range_channel_vals_arr = p_img_pixels_with_global_index_arr[:, channel_with_max_range_index_int+1]

		#----------
		# SORT_BY_CHANNEL
		# argsort() - sorts values and returns their indexes
		img_max_range_channel_vals_sorted_indexes_arr = img_max_range_channel_vals_arr.argsort()

		# split sorted list of pixels (on channel with max range by pixel value) indicies into half/median. 
		# np.array_split() - wont error if the image doesnt split in 2 equal parts.
		upper_half__px_indicies_arr, lower_half__px_indicies_arr = np.array_split(img_max_range_channel_vals_sorted_indexes_arr, 2)

		# index into image pixels by sorted indexes of the channel with biggest range/variance
		upper_half__px_arr = p_img_pixels_with_global_index_arr[upper_half__px_indicies_arr]
		lower_half__px_arr = p_img_pixels_with_global_index_arr[lower_half__px_indicies_arr]

		# RECURSION
		upper_half__avrg_pixels_arr = process_img__with_px_coords(upper_half__px_arr, p_level_int+1)
		lower_half__avrg_pixels_arr = process_img__with_px_coords(lower_half__px_arr, p_level_int+1)

		r=[]
		r.extend(upper_half__avrg_pixels_arr)
		r.extend(lower_half__avrg_pixels_arr)
		return r

	#----------------------------------------------

	# ADD_PIXEL_GLOBAL_INDEX - add an index of the pixel relative to the whole image, to the pixels rgb color value.
	#                          this is used to be able to track where in the image particular pixels come from.
	# [[r0, g0, b0], [r1, g1, b1], [r2, g2, b2], ...] -> [[0, r0, g0, b0], [1, r1, g1, b1], [2, r2, g2, b2], ...]
	# 
	# np.column_stack() - Stack 1-D arrays as columns into a 2-D array.
	img_pixels_with_global_index_arr = np.column_stack((np.arange(len(img_pixels_arr)), img_pixels_arr))


	start_level_int = 0
	result = process_img__with_px_coords(img_pixels_with_global_index_arr, start_level_int)


	# IMAGE_PALETTE
	r_img = Image.new('RGB', (len(result), 1))



	# get the palette pixels only, removing global pixel indexes from each of the palette pixel part.
	# each of the palette pixels has a format (px_global_indexes_arr, r, g, b), so only get (r, g, b).
	palette_pixels_only_arr = np.empty((len(result), 3), dtype=int)
	for i in range(len(result)):
		pixels_global_indexes_lst, r_int, g_int, b_int = result[i]
		palette_pixels_only_arr[i] = np.array([r_int, g_int, b_int])


		print(pixels_global_indexes_lst)


		print(f"global {len(result[i][0])} {palette_pixels_only_arr[i]}")

	print(palette_pixels_only_arr)
	print(palette_pixels_only_arr.shape)

	return palette_pixels_only_arr, result

#----------------------------------------------
def run_extended(p_image_file_path_str,
	p_palette__output_file_path_str):
 
	palette_pixels_only_arr, result = run_from_file(p_image_file_path_str)

	r_img = Image.fromarray(palette_pixels_only_arr.reshape(palette_pixels_only_arr.shape[0], 1, 3).astype(np.uint8))
	r_img.save(p_palette__output_file_path_str)


	#----------
	
	image = Image.open(p_image_file_path_str)
	width_int, height_int = image.size

	img_pixels_original_arr = np.array(image.getdata())

	# only select the first 3 channels of each pixel, since some images have
	# an alpha channel as well which we dont want.
	img_pixels_arr = img_pixels_original_arr[:, :3]

	# colorize original image with quantized color palette,
	# using the global_indexes of every average pixel color calcuated
	# to set those indexes to those exact values.
	# ADD!! - figure out how to vectorize this operation, withyout needing
	#         to iterate through pixels individually.
	colors_lst = []
	for c in result:

		global_indexes_arr = c[0]
		color_arr          = c[1:4]
		colors_lst.append(color_arr)

		# color_4d_arr = np.array(color_arr)
		# color_4d_arr = np.append(color_4d_arr, 1).reshape((4, ))

		for i in global_indexes_arr:
			img_pixels_arr[i] = color_arr

	# height/width/3 - height has to go before width
	img_pixels_3d_arr = img_pixels_arr.reshape(height_int, width_int, 3)

	r_img = Image.fromarray(img_pixels_3d_arr.astype(np.uint8))

	#----------------------------------------------
	# def draw_sectors():
	#     draw = ImageDraw.Draw(r_img)
	#     draw.rectangle(((0, 00), (100, 100)), outline='black', width=1)

	#----------------------------------------------
	# draw_sectors()

	r_img.save(f"{p_palette__output_file_path_str}__sectors.png")

	#----------

	print("K MEANS ------------")
	kmeans(colors_lst)

#----------------------------------------------
def kmeans(p_colors_lst, p_k = 5):

	import scipy.spatial.distance
	import matplotlib.pyplot as plt
	import sklearn.cluster

	# array([[ 25, 236, 102],
	#       [129, 147, 103],
	#       [226,  35,  64],
	#       [ 94,  15,  33],
	#       [ 56, 104, 115]], dtype=uint64)
	centroids_arr = (np.random.rand(p_k, 3)*255).round().astype(np.uint)

	print("------------------------")
	print(p_colors_lst)
	print("==")
	print(centroids_arr)

	#----------------------------------------------
	def algo():
		for i in range(0, 10):

			# ASSIGN_COLORS_TO_CENTROIDS
			colors_centroids_lst = []
			for color_tpl in p_colors_lst:
				color_to_centroids_distances_arr = scipy.spatial.distance.cdist(centroids_arr, [color_tpl], 'euclidean')

				# print("result:")
				# print(color_to_centroids_distances_arr)

				# get the index/label of the centroid that the color is closest to
				closest_centroid_index_int = color_to_centroids_distances_arr.argmin()
				colors_centroids_lst.append(closest_centroid_index_int)
			
			# GROUP_COLORS_BY_CENTROIDS
			centroids_colors_map = {}
			for i in range(0, len(colors_centroids_lst)):

				color_arr          = p_colors_lst[i]
				centroid_label_int = colors_centroids_lst[i]
				if centroid_label_int in centroids_colors_map.keys():
					centroids_colors_map[centroid_label_int].append(color_arr)
				else:
					centroids_colors_map[centroid_label_int] = [color_arr]



			for centroid_label_int, colors_lst in centroids_colors_map.items():

				colors_by_channel_arr = np.array(colors_lst).T
				new_r_f = np.average(colors_by_channel_arr[0])
				new_g_f = np.average(colors_by_channel_arr[1])
				new_b_f = np.average(colors_by_channel_arr[2])

				centroid_new_coords_arr = np.array([new_r_f, new_g_f, new_b_f])

				# print(centroid_new_coords_arr)

				centroids_arr[centroid_label_int] = centroid_new_coords_arr
		
		return colors_centroids_lst

	#----------------------------------------------
	def plot_colors(p_colors_centroids_lst):
		

		fig = plt.figure()
		ax1 = fig.add_subplot(1, 2, 1, projection="3d") # plt.axes(projection='3d')
		ax2 = fig.add_subplot(1, 2, 2, projection="3d")

		color_map_lst = [
			"red",
			"green",
			"blue",
			"yellow",
			"orange"
		]

		#----------------
		# CLUSTERS
		colors_rgb_coords_arr = np.array(p_colors_lst).T
		colors_x_arr = colors_rgb_coords_arr[0]
		colors_y_arr = colors_rgb_coords_arr[1]
		colors_z_arr = colors_rgb_coords_arr[2]

		colors_1range_arr = np.array(p_colors_lst)/255
		# ax.scatter3D(colors_x_arr, colors_y_arr, colors_z_arr, c=colors_1range_arr)

		colors_lst = [color_map_lst[l] for l in colors_centroids_lst]
		ax1.scatter3D(colors_x_arr, colors_y_arr, colors_z_arr, c=colors_lst)

		#----------------

		# CENTROIDS
		centroid_coords_arr = centroids_arr.T
		centroids_x_arr = centroid_coords_arr[0]
		centroids_y_arr = centroid_coords_arr[1]
		centroids_z_arr = centroid_coords_arr[2]
		ax1.scatter3D(centroids_x_arr, centroids_y_arr, centroids_z_arr, c="black")

		# plt.show()

		#----------------
		# KMEANS_SKLEARN		
		clusters_labels_lst = sklearn.cluster.KMeans(n_clusters=p_k).fit_predict(p_colors_lst)
		print(clusters_labels_lst)

		
		colors_lst = [color_map_lst[l] for l in clusters_labels_lst]
		ax2.scatter3D(colors_x_arr, colors_y_arr, colors_z_arr, c=colors_lst) # colors_1range_arr)

		#----------------

		plt.show()

	#----------------------------------------------
	colors_centroids_lst = algo()
	plot_colors(colors_centroids_lst)

#--------------------------------------------------
def parse_args():
	arg_parser = argparse.ArgumentParser(formatter_class = argparse.RawTextHelpFormatter)
	
	#----------------------------
	# INPUT_IMAGES_LOCAL_FILE_PATHS
	arg_parser.add_argument("-input_images_local_file_paths", action = "store", default=None,
		help = "list of image file paths (',' delimited) to process")

	#----------------------------
	# OUTPUT_DIR_PATH
	arg_parser.add_argument("-output_dir_path", action = "store", default=None,
		help = "dir path of the output images")

	#----------------------------
	# MEDIAN_CUT_LEVELS_NUM
	arg_parser.add_argument("-median_cut_levels_num", action = "store", default=None,
		help = "number of levels to use for the median-cut algo subdivisions")

	#----------------------------
	cli_args_lst   = sys.argv[1:]
	args_namespace = arg_parser.parse_args(cli_args_lst)

	return {
		"input_images_local_file_paths_str": args_namespace.input_images_local_file_paths,
		"output_dir_path_str":               args_namespace.output_dir_path,
		"median_cut_levels_num":             int(args_namespace.median_cut_levels_num)
	}

#----------------------------------------------


# GloFlow application and media management/publishing platform
# Copyright (C) 2021 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

# import subprocess
# import threading

from colored import fg, bg, attr

sys.path.append("%s/../utils"%(modd_str))
import gf_core_cli

#--------------------------------------------------
def run(p_test_ci_bool, p_aws_region_str):
	
	print(f"    {fg('green')}TEST MASTER_GET_BLOCK{attr(0)} >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")

	#--------------------------------------------------
	def start_master():
		bin_str = f"{modd_str}/../../build/gf_web3_monitor"
		cmd_lst = [
			bin_str,
			"start", "service",
			f"--config={modd_str}/../../config/gf_web3_monitor.yaml"
		]



		
		print(f'GF_SENTRY_ENDPOINT - {os.environ["GF_SENTRY_ENDPOINT"]}')
		print(f'GF_AWS_SQS_QUEUE   - {os.environ["GF_AWS_SQS_QUEUE"]}')

		# p = subprocess.Popen(cmd_lst, shell=False, stdout=subprocess.PIPE, bufsize=1,
		# 	env={
		# 		"AWS_REGION":            p_aws_region_str,
		# 		"AWS_ACCESS_KEY_ID":     os.environ["AWS_ACCESS_KEY_ID"],
		# 		"AWS_SECRET_ACCESS_KEY": os.environ["AWS_SECRET_ACCESS_KEY"],
		# 		"GF_SENTRY_ENDPOINT":    os.environ["GF_SENTRY_ENDPOINT"],
		# 		"GF_AWS_SQS_QUEUE":      os.environ["GF_AWS_SQS_QUEUE"],
		# 		"GF_EVENTS_CONSUME":        "false",
		# 		"GF_WORKERS_AWS_DISCOVERY": "false" # "true" # "false" # use the localy started worker_inspector, at test startup
		# 	})
		# t = threading.Thread(target=gf_test_utils.read_process_stdout, args=(p.stdout, "gf_eth_monitor", "magenta"))
		# t.start()


		p = gf_core_cli.run__view_realtime(cmd_lst, {
				"AWS_REGION":            p_aws_region_str,
				"AWS_ACCESS_KEY_ID":     os.environ["AWS_ACCESS_KEY_ID"],
				"AWS_SECRET_ACCESS_KEY": os.environ["AWS_SECRET_ACCESS_KEY"],
				"GF_SENTRY_ENDPOINT":    os.environ["GF_SENTRY_ENDPOINT"],
				"GF_AWS_SQS_QUEUE":      os.environ["GF_AWS_SQS_QUEUE"],

				"GF_EVENTS_CONSUME":        "false",
				"GF_WORKERS_AWS_DISCOVERY": "false" # "true" # "false" # use the localy started worker_inspector, at test startup
			},
			"gf_eth_monitor", "green")

		return p

	#--------------------------------------------------
	
	p = start_master()

	# give external services time to process, only in local test mode
	if not p_test_ci_bool:
		import time
		time.sleep(10)

	#--------------------------------------------------
	def test__master():
		import requests



		block_num_int = 10

		print("MAKING TEST CLIENT REQUEST")
		url_str = f"http://127.0.0.1:4050/gfethm/v1/block?b={block_num_int}"

		print(url_str)

		r = requests.get(url_str)


		print(r.text)


		import json
		r_map = json.loads(r.text)

		assert r_map["status_str"] == "OK"
		assert "data" in r_map.keys()
		assert "block_from_workers_map" in r_map["data"]
		assert "miners_map" in r_map["data"]
		assert isinstance(r_map["data"]["miners_map"], dict)
		
		for worker_host_str, block_info_map in r_map["data"]["block_from_workers_map"].items():
			assert isinstance(worker_host_str, str)
			assert isinstance(block_info_map, dict)


			print("BLOCK INFO ----------------")
			print(block_info_map)

			assert "block_num_int"     in block_info_map.keys()
			assert "gas_used_uint"     in block_info_map.keys()
			assert "gas_limit_uint"    in block_info_map.keys()
			assert "coinbase_addr_str" in block_info_map.keys()
			assert "txs_lst"           in block_info_map.keys()
			assert "txs_hashes_lst"    in block_info_map.keys()
			assert "time_uint"         in block_info_map.keys()
			assert "block"             in block_info_map.keys()
		



			# assert block_num_int == block_info_map["block_num_int"]


		print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> TEST COMPLETE ----------------")

	#--------------------------------------------------
	test__master()


	# give external services time to process, only in local test mode
	if not p_test_ci_bool:
		time.sleep(20)

	p.terminate()


# GloFlow application and media management/publishing platform
# Copyright (C) 2023 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
from urllib.parse import urlparse
import requests

#---------------------------------------------------------------------------------
def download_file(p_url_str, p_local_file_path_str):
    response = requests.get(p_url_str)
    with open(p_local_file_path_str, 'wb') as f:
        f.write(response.content)

#---------------------------------------------------------------------------------
def download_file_chunked(p_url_str,
    p_local_file_path_str):
    """Download an image from a URL and save it to the local filesystem."""
    
    # HTTP_GET
    response = requests.get(p_url_str, stream=True)

    if response.status_code == 200:
        
        # ensure the local directory exists
        if not os.path.dirname(p_local_file_path_str) == "":
            os.makedirs(os.path.dirname(p_local_file_path_str), exist_ok=True)

        # open the local file in binary write mode
        with open(p_local_file_path_str, 'wb') as file:

            for chunk in response.iter_content(1024):
                file.write(chunk)
        
        print(f"image downloaded: {p_local_file_path_str}")
    else:
        print(f"failed to download image. Status code: {response.status_code}")

#---------------------------------------------------------------------------------
def is_absolute_url(url):
    parsed_url = urlparse(url)
    return bool(parsed_url.netloc)


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

from colored import fg, bg, attr

sys.path.append("%s/../../gf_core"%(modd_str))
import gf_core_cli

#--------------------------------------------------
# RUST NIGHTLY
# rustup self update                 - rustup update
# rustup toolchain install nightly   - rustup install Rust nightly
# rustup run nightly rustc --version - rstup test Rust nightly
# rustup default nightly             - rustup make Rust nightly the global default

#--------------------------------------------------
# RUN_IN_CONTAINER
def run_in_cont():

    repo_local_path_str = os.path.abspath(f'{modd_str}/../../../../../gloflow').strip()
    cmd_lst = [
        "sudo", "docker", "run",
        "--rm", # remove after exit 
        "-v", f"{repo_local_path_str}:/home/gf", # mount repo into the container
        "glofloworg/gf_builder_rust_ubuntu:latest",

        # "python3", "-u", "/home/gf/build/gf_builder/gf_builder.py", "-run=build_rust"
        "python3", "-u", "/home/gf/ops/cli__build.py", "-run=build_rust", "-build_outof_cont", "-app=gf_images_jobs"
    ]
    p = gf_core_cli.run__view_realtime(cmd_lst, {},
        "gf_build_rust", "green")

    p.wait()

#--------------------------------------------------
# RUN
def run(p_cargo_crate_dir_path_str,
    p_static_bool       = False,
    p_exit_on_fail_bool = True,
    p_verbose_bool      = False):
    assert os.path.isdir(p_cargo_crate_dir_path_str)
    
    print(f"{fg('yellow')}BUILD{attr(0)}")
    print(f"crate dir - {fg('yellow')}{p_cargo_crate_dir_path_str}{attr(0)}")

    cwd_str = os.getcwd()
    os.chdir(os.path.abspath(p_cargo_crate_dir_path_str)) # change into the target main package dir

    #-------------
    # "rustup update stable"
    # _, _, exit_code_int = gf_cli_utils.run_cmd("cargo clean")

    #-------------

    c_lst = []

    if p_static_bool:

        # DOCUMENT!! - without this the py extension wont compile. 
        #              complaining that target for musl from gf_images_job lib cant be used, since this py extension
        #              package is marked as a dynamic lib (which it has to be to be importable by the Py VM).
        if os.path.basename(p_cargo_crate_dir_path_str) == "gf_images_jobs_py":
            c_lst.append("RUSTFLAGS='-C target-feature=-crt-static'")

    c_lst.extend([
        # 'RUSTFLAGS="$RUSTFLAGS -A warnings"', # turning off rustc warnings
        # "RUSTFLAGS='-L %s'"%(os.path.abspath("%s/../../rust/gf_images_jobs/test"%(modd_str))),

        # if compiling on Ubuntu for Alpine for example, this ENV var should be set
        # "PKG_CONFIG_ALLOW_CROSS=1",
        
        "cargo build",
    ])

    if p_verbose_bool:
        # c_lst.append("--verbose")
        c_lst.append("-vv") # very verbose


    # STATIC_LINKING - some outputed libs (imported by Go for example) should contain their
	#                  own versions of libs statically linked into them.
    if p_static_bool:

        #-------------
        # MUSL - staticaly compile libc compatible lib into the output binary. without MUSL
        #        rust statically compiles all program libs except the standard lib.
        # musl-gcc - musl-gcc is a wrapper around GCC that uses the musl C standard library
        #            implementation to build programs. It is well suited for being linked with other libraries
        #            into a single static executable with no shared dependencies.
        #            its used by "cargo build" if we target linux-musl.
        #            "sudo apt-get install musl-tools" - make sure its installed
        #
        # x86_64-unknown-linux-musl - for 64-bit Linux.
        #                             for this to work "rustup" has to be used to install this
        #                             build target into the Rust toolchain. 
        #                             (for GF CI this is done in the gf_builder Dockerfile__gf_builder)
        c_lst.append("--target x86_64-unknown-linux-musl")

        #-------------

    # DYNAMIC_LINKING
    else:
        c_lst.append("--release")


    cmd_str = " ".join(c_lst)
    print(cmd_str)
    _, _, exit_code_int = gf_core_cli.run(cmd_str)

    # IMPORTANT!! - if "go build" returns a non-zero exit code in some environments (CI) we
    #               want to fail with a non-zero exit code as well - this way other CI 
    #               programs will flag builds as failed.
    if not exit_code_int == 0:
        if p_exit_on_fail_bool:
            exit(exit_code_int)

    os.chdir(cwd_str) # return to initial dir

#--------------------------------------------------
def prepare_libs(p_name_str,
    p_cargo_crate_dir_path_str,
    p_type_str,
    p_exit_on_fail_bool = True):
    assert os.path.isdir(p_cargo_crate_dir_path_str)
    assert p_type_str == "lib_rust"


    print(f"{fg('yellow')}PREPARE LIBS{attr(0)}>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")

    

    target_build_dir_path_str = os.path.abspath(f"{modd_str}/../../../../rust/build")
    assert os.path.isdir(target_build_dir_path_str)

    target_lib_file_path_lst = []
    if p_type_str == "lib_rust":

        release_dir_str = f"{p_cargo_crate_dir_path_str}/target/release"
        gf_core_cli.run(f"ls -al {release_dir_str}")    

        #-------------
        # FIX!! - dont hardcode the app_name here like this, but parse Cargo.toml to detect if 
        #         one of the Crate types is "staticlib".
        if p_name_str == "gf_images_jobs":

            '''
            #-------------
            # EXTERN_LIB
            target_lib_dir_str = f"{modd_str}/../../../../rust/build" # f"{modd_str}/../../build/gf_apps/gf_images/tf_lib"

            prepare_libs__extern(target_lib_dir_str,
                p_tf_libs_bool=True,
                p_exit_on_fail_bool=p_exit_on_fail_bool)

            #-------------
            '''
            
            # RUST_PY - CPYTHON_EXTENSION - this lib is Python extension written in Rust.
            #                               at the moment in GF the convention is for these Rust libs to have a postfix "_py".
            if p_cargo_crate_dir_path_str.endswith("_py"):
                
                # DYNAMIC_LIB
                # IMPORTANT!! - Rust compiles this dynamic lib with the "lib" prefix, but the Python VM
                #               requires extension libs to not have the "lib" prefix.
                source__py_lib_file_path_str = f"{release_dir_str}/lib{p_name_str}_py.so"
                target__py_lib_file_path_str = f"{target_build_dir_path_str}/{p_name_str}_py.so"

                assert os.path.isfile(source__py_lib_file_path_str)
                target_lib_file_path_lst.append((source__py_lib_file_path_str, target__py_lib_file_path_str))

            else:

                # STATIC_LIB
                source__static_lib_file_path_str = f"{release_dir_str}/lib{p_name_str}.a"
                assert os.path.isfile(source__static_lib_file_path_str)
                target_lib_file_path_lst.append((source__static_lib_file_path_str, target_build_dir_path_str))

                # DYNAMIC_LIB
                source__dynamic_lib_file_path_str = f"{release_dir_str}/lib{p_name_str}.so"
                assert os.path.isfile(source__dynamic_lib_file_path_str)
                target_lib_file_path_lst.append((source__dynamic_lib_file_path_str, target_build_dir_path_str))

        #-------------
        # ALL
        else:
            target_lib_file_path_lst.append((f"{release_dir_str}/lib{p_name_str}.so"), target_build_dir_path_str)

        #-------------


    # COPY_FILES
    for source_f, target_f in target_lib_file_path_lst:
        c_str = "cp %s %s"%(source_f, target_f)
        _, _, exit_code_int = gf_core_cli.run(c_str) # gf_cli_utils.run_cmd(c_str)



    # IMPORTANT!! - if "go build" returns a non-zero exit code in some environments (CI) we
    #               want to fail with a non-zero exit code as well - this way other CI 
    #               programs will flag builds as failed.
    if not exit_code_int == 0:
        if p_exit_on_fail_bool:
            exit(exit_code_int)
    

#--------------------------------------------------
'''
def prepare_libs__extern(p_target_lib_dir_str,
    p_tf_libs_bool      = False,
    p_exit_on_fail_bool = True):

    #-------------
    # TENSORFLOW
    # IMPORTANT!! - download TF lib and place it in appropriate dir, to have a fresh TF libs
    #               in the build server context, without including the lib in the repo itself.
    if p_tf_libs_bool:

        print(f"{fg('green')}prepare TensorFlow lib{attr(0)}")

        lib_file_name_str  = "tflib.tar.gz"
        tf__version_str    = "1.15.0"
        tf__filename_str   = f"libtensorflow-cpu-linux-x86_64-{tf__version_str}.tar.gz"
        tf__url_str        = f"https://storage.googleapis.com/tensorflow/libtensorflow/{tf__filename_str}"

        gf_core_cli.run(f"mkdir -p {p_target_lib_dir_str}/tf_lib")

        # DOWNLOAD
        _, _, exit_code_int = gf_core_cli.run(f"curl {tf__url_str} --output {lib_file_name_str}")
        if not exit_code_int == 0:
            if p_exit_on_fail_bool:
                exit(exit_code_int)

        # UNPACK
        gf_core_cli.run(f"mv {lib_file_name_str} {p_target_lib_dir_str}/{lib_file_name_str}")
        gf_core_cli.run(f"tar -xvzf {p_target_lib_dir_str}/{lib_file_name_str} -C {p_target_lib_dir_str}/tf_lib")

    #-------------
'''


# GloFlow application and media management/publishing platform
# Copyright (C) 2023 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
from PIL import Image

import gf_color_palette

#----------------------------------------------
def main(p_test_bool=True):

	# TEST
	if p_test_bool:
		test_image_paths_lst = [
			'./../../gf_ml_worker/test/data/input/1234cd19517b939d3eb726c817985fe4_thumb_medium.jpeg',
			'./../../gf_ml_worker/test/data/input/canvas.png',
			"./../../gf_ml_worker/test/data/input/4b14ca75070ac78323cf2ddef077ae92_thumb_medium.jpeg",
			"./../../gf_ml_worker/test/data/input/4838df39722bc2d681b67bf739f29357_thumb_small.jpeg",
			"./../../gf_ml_worker/test/data/input/1234cd19517b939d3eb726c817985fe4_thumb_medium.jpeg",
			"./../../gf_ml_worker/test/data/input/3a61e7d68fb17198e8dc0476cc862ddd_thumb_small.jpeg",
		]


		for i in range(0, len(test_image_paths_lst)):

			image_file_path_str = test_image_paths_lst[i]
			o_str               = f"./output/out_{i}.png"

			# RUN
			gf_color_palette.run_extended(image_file_path_str, o_str)

	# PRODUCTION
	else:

		args_map = parse_args()

		input_images_local_file_paths_lst = args_map["input_images_local_file_paths_str"].split(",")
		
		# RUN
		run_multiple(input_images_local_file_paths_lst)

		out_map = {}
		print(f"GF_OUT:{json.dumps(out_map)}")

#----------------------------------------------
def run_multiple(p_input_images_local_file_paths_lst):

	print(f"RUNNING {fg('green')}GF_COLOR_PALETTE{attr(0)} PLUGIN ")
	
	# VERIFY
	for f in p_input_images_local_file_paths_lst:
		assert os.path.isfile(f)
		print(f"input - {fg('yellow')}{f}{attr(0)}")

	# RUN
	for f in p_input_images_local_file_paths_lst:
		gf_color_palette.run_extended(f)

#----------------------------------------------
if __name__ == "__main__":
	main()


# GloFlow application and media management/publishing platform
# Copyright (C) 2020 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__)) # module dir

import subprocess

#-------------------------------------------------------------
# SERVICES__CATALOG
def services__catalog(p_auth_token_str,
	p_ca_intermediate__file_path_str = None,
	p_cert_combined__file_path_str   = None,
	p_cert_key__file_path_str        = None,
	p_host_str                       = "127.0.0.1:8501"):

	# cmd_lst = []
	# if p_sudo_bool:
	# 	cmd_lst.append("sudo")
	#
	# cmd_lst.extend([
	# 	"CONSUL_HTTP_TOKEN=%s"%(p_auth_token_str),
	# 	"consul catalog services",
	# 	"-tags",
	# 	"-http-addr=https://%s"%(p_host_str),
	# ])
	
	# CONSUL_HTTP_API
	assert os.path.isfile(p_ca_intermediate__file_path_str)
	assert os.path.isfile(p_cert_combined__file_path_str)
	assert os.path.isfile(p_cert_key__file_path_str)

	url_str = "https://%s/v1/catalog/services"%(p_host_str)
	r = requests.get(url_str,
		verify=p_ca_intermediate__file_path_str,
		cert=(p_cert_combined__file_path_str, p_cert_key__file_path_str))


	print(r.text)
	
#-------------------------------------------------------------
# AGENT__START
def agent__start(p_name_str,
	p_type_str                 = "server",
	p_container_image_name_str = "consul:1.7.2",
	p_config__file_path_str    = None,
	p_external_ip_str          = "127.0.0.1",
	p_root_agent_ip_str        = "127.0.0.1",
	p_bootstrap_server_bool    = False,
	p_datacenter_str           = "us-east-1",
	p_volume_mounts_lst        = None,
	p_sudo_bool                = False):
	assert p_type_str == "client" or p_type_str == "server"
	assert os.path.isfile(p_config__file_path_str)

	cmd_lst = []
	if p_sudo_bool:
		cmd_lst.append("sudo")

	cmd_lst.extend([
		"docker run",
		"--restart=always",
		"--net=host",
		"--name=%s"%(p_name_str),

		# turn off the Consul dashboard
		'''-e "CONSUL_UI=false"'''
	])

	#-------------
	# VOLUMES
	if not p_config__file_path_str == None:
		cmd_lst.append("-v %s:/consul/config/%s"%(p_config__file_path_str, os.path.basename(p_config__file_path_str)))

	if not p_volume_mounts_lst == None:
		assert isinstance(p_volume_mounts_lst, list)
		for source_str, target_str in p_volume_mounts_lst:
			cmd_lst.append("-v %s:%s"%(source_str, target_str))

	#-------------

	if p_type_str == "client":

		# ENV
		cmd_lst.extend([
			# '''-e "CONSUL_LOCAL_CONFIG={'leave_on_terminate': true}"''' # its default since Consul 0.7
		])
		
	# MAIN
	cmd_lst.extend([
		p_container_image_name_str,

		"agent",
		"-datacenter=%s"%(p_datacenter_str),
	])

	if p_type_str == "server":
		cmd_lst.extend([

			# BIND_IP
			# this is the address advertised to the rest of the cluster.
			# should be reachable by all other nodes in the cluster.
			# Consul uses both TCP and UDP and the same port for both.
			"-bind=%s"%(p_external_ip_str),
			
			# specifies how many server agents to watch for before bootstrapping the cluster for the first time
			"-bootstrap-expect=1"
		])

		if p_bootstrap_server_bool:

			# flag is used to control if a server is in "bootstrap" mode. It is important that
			# no more than one server per datacenter be running in this mode. Technically, a server in 
			# bootstrap mode is allowed to self-elect as the Raft leader.
			cmd_lst.append("-bootstrap")

	if p_type_str == "client":

		cmd_lst.extend([
			# address of the agent to use to join the cluster
			"-retry-join=%s"%(p_root_agent_ip_str),

			# DNS
			# "-dns-port=53"
		])

	c_str = " ".join(cmd_lst)
	print(c_str)
	consul_agent_p = subprocess.Popen(c_str, shell=True, stdout=subprocess.PIPE, bufsize=1)
	
	return consul_agent_p


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import json
import urllib
import requests

default_host_str = 'https://gloflow.com'

#---------------------------------------------------
# sends the HTTP request to the gf_images service, to process the image
# ->:Map(image_processing_results_map)

def add_image(p_image_url_str,
	p_log_fun  = None,
	p_host_str = default_host_str,
	p_process_from_scratch_if_prexisting_bool = True):
	
	assert p_host_str.startswith('http://') or \
		p_host_str.startswith('https://')
				 
	# new_image_url_str = base64.b64encode(p_image_url_str)
	new_image_url_str = urllib.quote(p_image_url_str)
	url_str           = f'{p_host_str}/images/jobs/start'
	headers_map = {
		'accept': 'text/event-stream' # use Server Side Events
	}
	params_map  = {
		'type':    'process_extern_image',
		'img_url': new_image_url_str
	}
	
	# params_map  = {
	# 	'image_url_str':                           p_image_url_str,
	# 	'process_from_scratch_if_prexisting_bool': p_process_from_scratch_if_prexisting_str
	# }

	if not p_log_fun == None:
		p_log_fun('INFO', 'url_str - [%s]'%(url_str))

	r = requests.get(url_str,
		params  = params_map,
		headers = headers_map,
		stream  = True) # HTTP Server-Side-Events
	assert r.status_code == 200

	'''
	#---------------------------------------------------
	# SSE - streaming data lines

	# ->:List<:Map>
	def stream_responses():
		
		data_items_lst = []
		for line_str in r.iter_lines():

			if not p_log_fun == None:
				p_log_fun('INFO', '>>>>>>>>>>>>>>>>>>>>>>>>')
				p_log_fun('INFO', line_str)

			# filter out keep-alive new lines
			if line_str:
				if line_str.startswith('data: '):
					msg_str  = line_str.strip('data: ')
					msg_map = json.loads(msg_str)

					assert msg_map.has_key('status_str')
					status_str = msg_map['status_str']

					assert isinstance(status_str, str)
					assert status_str == 'ok' or \
						status_str == 'error'
					assert status_str == 'ok'

					data_map = msg_map['data_map']
					assert isinstance(data_map, dict)

					data_items_lst.append(data_map)
		return data_items_lst

	#---------------------------------------------------
	data_items_lst = stream_responses()
	'''



	resp_map = json.loads(r.text)
	assert(isinstance(resp_map.has_key("status"), str))
	assert(isinstance(resp_map["data"], dict))


	data_map = resp_map["data"]

	running_job_id_str = data_map["running_job_id_str"]
	job_expected_outputs_lst = data_map["job_expected_outputs_lst"]

	assert isinstance(job_expected_outputs_lst, list)
	assert len(job_expected_outputs_lst) == 1

	new_image_info_map = job_expected_outputs_lst[0]
	assert isinstance(new_image_info_map, dict)




	assert(p_image_url_str == new_image_info_map["image_source_url_str"])

	# IMAGE_ID
	assert "image_id_str" in new_image_info_map.keys()
	image_id_str = new_image_info_map['image_id_str']
	assert isinstance(image_id_str, str)




	#----------------
	# THUMBS

	assert new_image_info_map.has_key('thumbnail_small_relative_url_str')
	thumbnail_small_relative_url_str = new_image_info_map['thumbnail_small_relative_url_str']
	assert isinstance(thumbnail_small_relative_url_str, str)
	assert thumbnail_small_relative_url_str.startswith('/images')
	
	assert new_image_info_map.has_key('thumbnail_medium_relative_url_str')
	thumbnail_medium_relative_url_str = new_image_info_map['thumbnail_medium_relative_url_str']
	assert isinstance(thumbnail_medium_relative_url_str, str)
	assert thumbnail_medium_relative_url_str.startswith('/images')
	
	assert new_image_info_map.has_key('thumbnail_large_relative_url_str')
	thumbnail_large_relative_url_str = new_image_info_map['thumbnail_large_relative_url_str']
	assert isinstance(thumbnail_large_relative_url_str, str)
	assert thumbnail_large_relative_url_str.startswith('/images')
	
	#----------------

	image_results_map = {
		'image_id_str': image_id_str,
		'thumbs_info_map': {
 			'thumbnail_small_relative_url_str':  thumbnail_small_relative_url_str,
			'thumbnail_medium_relative_url_str': thumbnail_medium_relative_url_str,
			'thumbnail_large_relative_url_str':  thumbnail_large_relative_url_str
		}
	}
	
	return image_results_map


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys
modd_str = os.path.abspath(os.path.dirname(__file__))

import pandas as pd
import matplotlib.pyplot as plt 
import boto3
from colored import fg, bg, attr

sys.path.append("%s/../../../gf_aws"%(modd_str))
import gf_aws_s3

#---------------------------------------------------
def stats__image_buckets_general(p_aws_access_key_id_str,
	p_aws_secret_access_key_str):

	gf_s3_info = gf_aws_s3.s3_connect(p_aws_access_key_id_str, p_aws_secret_access_key_str)
	assert isinstance(gf_s3_info, gf_aws_s3.Gf_s3_info)

	# LIST_BUCKETS
	for bucket in gf_s3_info.s3_resource.buckets.all():
		print(bucket.name)

	#-------------------
	main_images__bucket_info_map    = process_bucket("gf--img",             gf_s3_info.imgs__bucket)
	crawler_images__bucket_info_map = process_bucket("gf--discovered--img", gf_s3_info.discovered_imgs__bucket)

	#-------------------
	
	view_bucket_info([crawler_images__bucket_info_map, main_images__bucket_info_map])

#---------------------------------------------------
def process_bucket(p_name_str, p_bucket):

	i                        = 0
	total_size_kb_int        = 0
	counts_per_day_map       = {}
	counts_per_file_type_map = {
		"thumbnails/":     0, # used??
		"thumbnails":      0,
		"thumbnails_jpeg": 0,
		"thumbnails_png":  0
	}

	for o in p_bucket.objects.all():
		file_name_str       = o.key
		dt                  = o.last_modified
		file_size_bytes_int = o.size
		file_size_kb_int    = file_size_bytes_int/1024

		#IMPORTANT!! - AWS access logs are stored in a /logs dir in some of the buckets
		if 'logs' in file_name_str:
			continue

		#-------------------
		# COUNTS_PER_EXTENSION

		file_ext_str = file_name_str.split('.')[-1:][0].lower()

		if file_ext_str == "jpg":
			file_ext_str = "jpeg"

		# FILE_EXTENSION_COUNTS
		if file_ext_str in counts_per_file_type_map.keys():
			counts_per_file_type_map[file_ext_str] += 1
		else:
			counts_per_file_type_map[file_ext_str] = 1

		# THUMBNAILS
		if "thumbnails/" in file_name_str:
			counts_per_file_type_map["thumbnails"] += 1

			if file_ext_str == "jpeg":  counts_per_file_type_map["thumbnails_jpeg"] += 1
			elif file_ext_str == "png": counts_per_file_type_map["thumbnails_png"] += 1

		#-------------------
		#TOTAL_FILE_COUNTS_PER_DAY

		day_str = '%s-%02d-%02d'%(dt.year,dt.month,dt.day)
		if day_str in counts_per_day_map.keys():
			counts_per_day_map[day_str] += 1
		else:
			counts_per_day_map[day_str] = 0

		#-------------------

		print("%s - %s - %sKB")%(i,file_name_str,file_size_kb_int)

		total_size_kb_int += file_size_kb_int
		i+=1

	return {
		"name_str":                 p_name_str,
		"total_size_kb_int":        total_size_kb_int,
		"counts_per_day_map":       counts_per_day_map,
		"counts_per_file_type_map": counts_per_file_type_map,
	}

#---------------------------------------------------
def view_bucket_info(p_infos_lst):

	#-------------------------
	for info_map in p_infos_lst:
		print("")
		print("")
		print("")
		print('bucket_name_str - %s'%(info_map['name_str']))
		print("")

		sorted_lst = info_map['counts_per_day_map'].items()
		sorted_lst.sort()
		
		for k,v in sorted_lst:
			print('%s -- %s'%(k, v)) 

		print("")
		print("")
		print("")
		for k,v in info_map['counts_per_file_type_map'].items():
			print('%s%s%s -- %s'%(bg('blue'), k, attr(0), v))

		total_size_mb_int = info_map['total_size_kb_int']/1024
		print('total size - %s%s%smb'%(bg('green'), total_size_mb_int, attr(0)))
		print("")
		print("")
		print("")

	#-------------------------

	df_lst = []
	for bucket_info_map in p_infos_lst:

		day_counts_sorted_lst = sorted([(d_str,c_int) for d_str,c_int in bucket_info_map['counts_per_day_map'].items()])

		# days_lst   = []
		# counts_lst = []
		for d, c in day_counts_sorted_lst:

			# all_buckets__days_set.add(d)
			# days_lst.append(d)
			# counts_lst.append(c)

			bname_str = bucket_info_map['name_str']
			d = {
				'day':                   d,
				'bucket':                bname_str,
				'%s__count'%(bname_str): c
			}
			df_lst.append(d)


	# df_columns_map = {
	# 	"days":        list(all_buckets__days_set),
	# 	"total_counts":counts_lst,
	# }
	# df = pd.DataFrame(data=df_columns_map)

	df = pd.DataFrame(df_lst)
	df.set_index("day", drop=True, inplace=True)
	df = df.fillna(0)
	df = df.sort_values(by=['day'])
	#print df

	#-------------------------
	# PLOT

	df.plot.line(figsize=(10, 6), alpha=0.75)

	# fig = plt.figure(figsize=(30,10))

	plt.title("S3 buckets info per day", fontsize=18)
	plt.xlabel("day",                    fontsize=14)
	plt.ylabel('# of images',            fontsize=14)
	plt.xticks(size = 6)
	plt.axes().yaxis.grid() # horizontal-grid

	plt.show()

	#-------------------------


# GloFlow application and media management/publishing platform
# Copyright (C) 2019 Ivan Trajkovic
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import traceback
import pprint
import sentry_sdk
from colored import fg, attr

#--------------------------------------------------------
def create(p_msg_str,
	p_type_str,
	p_data_map,
	p_exception,
	p_sybsystem_name_str,
	p_log_fun,
	p_sentry_bool=True,
	p_reraise_bool=False):
	assert isinstance(p_msg_str, str)
	assert isinstance(p_type_str, str)
	assert isinstance(p_data_map, dict) or p_data_map == None
	assert isinstance(p_exception, Exception)
	assert isinstance(p_sybsystem_name_str, str)
	
	full_msg_str = f'''
		{fg('green')}{p_msg_str}{attr('reset')}
		type:           {fg('green')}{p_type_str}{attr('reset')}
		subsys:         {fg('green')}{p_sybsystem_name_str}{attr('reset')}
		exception args: {fg('green')}{p_exception.args}{attr('reset')}
		data:           {fg('green')}{pprint.pformat(p_data_map)}{attr('reset')}
		trace: {fg('green')}{traceback.format_exc()}{attr('reset')}
	'''
	p_log_fun('ERROR', full_msg_str)

	if p_sentry_bool:
		sentry_sdk.capture_exception(p_exception)
	
	if p_reraise_bool:
		raise p_exception